__author__ = "PedroCR"

import numpy as np
import logging
from lifeActuary.mortality_table import MortalityTable


class MortalityTable2Heads:
    """
    Instantiates a mortality table for 2 heads, considering survival independence.
    """

    def __new__(cls, mt_old, mt_young, age_dif, status):
        all_status = {'joint-life', 'last-survivor'}
        bool_mt_old = isinstance(mt_old, MortalityTable)
        bool_mt_young = isinstance(mt_young, MortalityTable)
        bool_status = False
        if status in all_status:
            bool_status = True
        if bool_mt_old and bool_mt_young and bool_status:
            return object.__new__(cls)
        else:
            logging.warning(f'We need two instances of mortality tables and a possible status; {all_status}.')
            return None

    def __init__(self, mt_old, mt_young, age_dif=0, status='joint-life'):
        self.__mt_old = mt_old
        self.__mt_young = mt_young
        self.__age_dif = age_dif
        self.__status = status

        self.__mt_2heads = self.__create_mortality_table_2heads__()

    def __repr__(self):
        return f"{self.__class__.__name__}{self.mt_old, self.mt_young, self.age_dif, self.status}"

    @property
    def mt_old(self):
        return self.__mt_old

    @property
    def mt_young(self):
        return self.__mt_young

    @property
    def age_dif(self):
        return self.__age_dif

    @property
    def status(self):
        return self.__status

    @property
    def mt_2heads(self):
        return self.__mt_2heads

    # create the mortality table for the 2 heads
    def __create_mortality_table_2heads__(self):
        l0 = max(self.mt_old.lx[0], self.mt_young.lx[0])
        if self.status == 'joint-life':
            w = min(self.mt_old.w - self.age_dif, self.mt_young.w)
            qx = [1 - self.mt_old.npx(x=x + self.age_dif, n=1) * self.mt_young.npx(x=x, n=1)
                  for x in range(w + 1)]
        if self.status == 'last-survivor':
            w = max(self.mt_old.w - self.age_dif, self.mt_young.w)
            qx = [self.mt_old.nqx(x=x + self.age_dif, n=1) * self.mt_young.nqx(x=x, n=1)
                  for x in range(w + 1)]
        qx.insert(0, 0)  # appends the minimum age in the table
        return MortalityTable(mt=qx)

    # Probabilities
    def nqxy(self, x_young, x_old, n=1, method='udd'):
        """

        :param x_young:
        :param x_old:
        :param n:
        :param method:
        :return:
        """
        if x_young > x_old:
            return np.nan
        if self.status == 'joint-life':
            return 1 - self.mt_young.npx(x_young, n, method) * self.mt_old.npx(x_old, n, method)
        if self.status == 'last-survivor':
            return self.mt_young.nqx(x_young, n, method) * self.mt_old.nqx(x_old, n, method)

    def npxy(self, x_young, x_old, n=1, method='udd'):
        """

        :param x_young:
        :param x_old:
        :param n:
        :param method:
        :return:
        """
        if x_young > x_old:
            return np.nan
        if self.status == 'joint-life':
            return self.mt_young.npx(x_young, n, method) * self.mt_old.npx(x_old, n, method)
        if self.status == 'last-survivor':
            return 1 - self.mt_young.nqx(x_young, n, method) * self.mt_old.nqx(x_old, n, method)

    def t_nqxy(self, x_young, x_old, t=1, n=1, method='udd'):
        """

        :param x_young:
        :param x_old:
        :param t:
        :param n:
        :param method:
        :return:
        """
        if x_young > x_old:
            return np.nan
        return self.nqxy(x_young, x_old, t + n, method) - self.nqxy(x_young, x_old, t, method)

    # Benefits
    ## Annuities

    def annuity(self, x_young, x_first_payment, x_last_payment, i, g=0, m=1, method='udd'):
        """

        :param x_young:
        :param x_first_payment:
        :param x_last_payment:
        :param i:
        :param g:
        :param m:
        :param method:
        :return:
        """
        if x_young > self.mt_2heads.w or x_young > x_first_payment:
            return .0
        if x_young > self.mt_2heads.w + 1:
            return .0
        if int(m) != m:
            return np.nan
        v = 1 / (1 + i / 100)
        vg = (1 + g / 100) * v
        if x_last_payment > self.mt_2heads.w + 1:
            x_last_payment = self.mt_2heads.w + 1
        number_of_payments = int((x_last_payment - x_first_payment) * m) + 1
        x_lp = x_first_payment + (number_of_payments - 1) / m
        payments_instants = np.linspace(x_first_payment - x_young, x_lp - x_young, number_of_payments)
        terms = [self.npxy(x_young, x_young + self.age_dif, t, method) * vg ** t for t in payments_instants]
        return np.sum(terms) / (1 + g / 100) ** (x_first_payment - x_young) / m


    def naaxy(self, x, y, n, i, g, m):
        pass


import numpy as np


class Annuities_Certain:

    def __new__(cls, interest_rate, m):
        if interest_rate < 0 or m < 0 or int(m) != m:
            print(f"We need a rate of interest non negative and a positive integer frequency")
            return None
        return object.__new__(cls)

    def __init__(self, interest_rate, m=1):
        """
        This class instantiates the methods for the computation of financial annuities, for a given
        interest rate and a chosen frequency of payments m (in each period of the interest rate)

        :param interest_rate: interest rate, in percentage (e.g. use 5 for 5%)
        :param m: frequency of payments, in each period of the interest rate

        :return: several methods and annuities for the interest rate and frequency.
        """
        self.interest_rate = interest_rate / 100.
        self.frequency = m

        self.v = 1 / (1 + self.interest_rate)
        self.im = self.frequency * (np.power(1 + self.interest_rate, 1 / self.frequency) - 1)
        self.vm = np.power((1 + self.im / self.frequency), -1)
        self.dm = self.im * self.vm

    def check_terms(func):
        def func_wrapper(self, terms, *args, **kwargs):
            if not terms:
                terms=0
            if terms < 0 or int(terms) != terms:
                return np.nan
            res = func(self, terms)
            return res

        return func_wrapper

    def check_grow(func):
        def func_wrapper(self, terms, payment, grow):
            if grow / 100 <= -1 or terms < 0 or int(terms) != terms:
                return np.nan
            res = func(self, terms, payment, grow)
            return res

        return func_wrapper

    # Constant Term Financial Annuities

    @check_terms
    def an(self, terms):
        """
        Returns the present value of an immediate n-term financial annuity with payments equal to 1.
        Payments are made in the end of the periods. In fractional annuities, payments of 1/m are made
        m times per year at the end of the periods.

        :param terms: number of years

        :return: Expected Present Value (EPV) of an immediate n-term financial annuity
        """
        if not terms:
            return 1 / self.im
        return (1 - np.power(self.vm, terms * self.frequency)) / self.im

    @check_terms
    def aan(self, terms):
        """
        Returns the present value of a due n-term financial annuity with payments equal to 1.
        Payments are made in the end of the periods. In fractional annuities, payments of 1/m are made
        m times per year at the end of the periods.

        :param terms: number of years

        :return: Expected Present Value (EPV) of a due n-term financial annuity
        """
        if not terms:
            return 1 / self.dm
        return (1 - np.power(self.vm, terms * self.frequency)) / self.dm

    # Variable Terms Financial Annuities
    @check_terms
    def Ian(self, terms, payment=1, increase=1):
        """
        Returns the present value of an immediate $n$ term financial annuity with payments
        increasing/decreasing arithmetically. Payments are made in the end of the periods.
        First payment and increase amount may differ.
        In fractional annuities, payments level within each interest period and increase/decrease from one
        interest period to the next.

        :param terms: number of years
        :param payment: amount of the first payment
        :param increase: increase amount of payments (positive for increasing annuities and negative for decreasing annuities)

        :return: Expected Present Value (EPV) of an immediate arithmetically increasing/decreasing financial annuity. Payments level within each interest period and increase/decrease from one interest period to the next.
        """
        if payment + increase * terms < 0:
            return np.nan
        # (payment - increase) * self.an(terms) + increase * (self.aan(terms) - terms * self.v ** terms) / self.im
        return payment * self.an(terms) + increase / self.im * (
                    (1 - self.v ** terms) / self.interest_rate - terms * self.v ** terms)

    @check_terms
    def Iaan(self, terms, payment=1, increase=1):
        """
        Returns the present value of a due $n$ term financial annuity with payments increasing/decreasing arithmetically. Payments are made in the beginning of the periods.
        First payment and increase amount may differ.
        In fractional annuities, payments level within each interest period and increase/decrease from one
        interest period to the next.

        :param terms: number of years
        :param payment: amount of the first payment
        :param increase: increase amount of payments (positive for increasing annuities and negative for decreasing annuities)

        :return: Expected Present Value (EPV) of a due arithmetically increasing/decreasing financial annuity. Payments level within each interest period and increase/decrease from one interest period to the next
        """
        return self.Ian(terms, payment, increase) / self.vm

    @check_terms
    def Iman(self, terms, payment=1, increase=1):
        """
        Returns the present value of an immediate $n$-term financial annuity with payments
        increasing/decreasing arithmetically.
        Payments are made in the end of the periods.
        First payment and increase amount may differ.
        In fractional annuities, payments increase in each payment period.

        :param terms: number of years
        :param payment: amount of the first payment
        :param increase: increase amount of payments (positive for increasing annuities and negative for decreasing annuities)

        :return: Expected Present Value (EPV) of an arithmetically increasing/decreasing financial annuity.
        Payments increase in each payment period and are paid in the end of periods.
        """
        if payment + increase * terms < 0:
            return np.nan

        return (payment - increase) * self.an(terms) \
               + increase * self.v \
               * (self.v ** terms * ((terms * self.frequency) * (self.vm - 1) - 1) + 1) \
               / (self.frequency * self.v ** ((self.frequency - 1) / self.frequency) * (self.vm - 1) ** 2)

    @check_terms
    def Imaan(self, terms, payment=1, increase=1):
        """
        Returns the present value of an immediate $n$-term financial annuity with payments
        increasing/decreasing arithmetically.
        Payments are made in the beginning of the periods.
        First payment and increase amount may differ.
        In fractional annuities, payments increase in each payment period.

        :param terms: number of years
        :param payment: amount of the first payment
        :param increase: increase amount of payments (positive for increasing annuities and negative for decreasing annuities)

        :return: Expected Present Value (EPV) of an arithmetically increasing/decreasing financial annuity.
        Payments increase in each payment period and are paid in the beginning of periods.
        """

        return self.Iman(terms, payment, increase) / self.vm

    @check_grow
    def Gan(self, terms, payment=1, grow=0):
        """
        Returns the present value of an immediate $n$ term financial annuity with payments increasing/decreasing geometrically. Payments are made in the end of the periods. In fractional annuities, payments level within each interest period and increase/decrease from one interest period to the next.

        :param terms: number of years
        :param payment: amount of the first payment
        :param grow: growth rate from one year to the other

        :return: Expected Present Value (EPV) of an arithmetically increasing/decreasing financial annuity. Payments increase in each year.
        """

        v = (1 + grow / 100) * self.v
        if self.interest_rate == grow / 100:
            return payment * terms * self.frequency * self.vm / self.frequency
        return payment / (1 + grow / 100) ** (1 / self.frequency) * (1 - v ** terms) / \
               (1 - v ** (1 / self.frequency)) * v ** (1 / self.frequency) / self.frequency

    @check_grow
    def Gaan(self, terms, payment=1, grow=0):
        """
        Returns the present value of a due $n$ term financial annuity with payments increasing/decreasing geometrically. Payments are made in the beginning of the periods. In fractional annuities, payments level within each interest period and increase/decrease from one interest period to the next.

        :param terms: number of years
        :param payment: amount of the first payment
        :param grow: growth rate from one year to the other

        :return: Present Value of an arithmetically increasing/decreasing financial annuity. Payments increase in each year.
        """
        v = (1 + grow / 100) * self.v
        return self.Gan(terms, payment, grow) / v ** (1 / self.frequency)

    @check_grow
    def Gman(self, terms, payment=1, grow=0):
        """
        Returns the present value of an immediate $n$ term financial annuity with payments increasing/decreasing geometrically. Payments are made in the end of the periods. In fractional annuities, payments increase in each payment period.

        :param terms: number of years
        :param payment: amount of the first payment
        :param grow: rate of growing of payments, in percentage

        :return: Present Value of an arithmetically increasing/decreasing financial annuity with terms paid in the end of periods.
        Payments increase in each payment period.
        """
        a1 = (1 - self.v) / self.im
        if self.interest_rate == grow / 100:
            return a1 * terms
        ig = (self.interest_rate - grow / 100) / (1 + grow / 100)
        vg = 1 / (1 + ig)
        a2 = (1 - vg ** terms) / (1 - vg)
        return payment * a1 * a2

    @check_grow
    def Gmaan(self, terms, payment=1, grow=0):
        """
        Returns the present value of an immediate $n$ term financial annuity with payments increasing/decreasing geometrically. Payments are made in the beginning of the periods. In fractional annuities, payments increase in each payment period.

        :param terms: number of years
        :param payment: amount of the first payment
        :param grow: rate of growing of payments, in percentage

        :return: Present Value of an arithmetically increasing/decreasing financial annuity with terms paid in the beginning of periods.
        Payments increase in each payment period.
        """
        v = (1 + grow / 100) * self.v
        return self.Gman(terms, payment, grow) / v ** (1 / self.frequency)



__author__ = "PedroCR"

import numpy as np
import logging
from lifeActuary import mortality_table as mt
from lifeActuary import mortality_table_2heads as mt2h
from lifeActuary import commutation_table as ct
from lifeActuary import mortality_insurance as mi

all_status = {'joint-life', 'last-survivor'}
log_message_mt = 'We need mtx and mty to be an instance of the class MortalityTable.'
log_message_status = f'Please, check the status. Status available are {all_status}'


def check_mortality_tables(func):
    def wrapper(*args, **kwargs):
        mt_in_args = [isinstance(m, mt.MortalityTable) for m in args[0:2]]
        mt_in_kwargs = [isinstance(m, mt.MortalityTable) for m in kwargs.values()]
        if sum(mt_in_args) + sum(mt_in_kwargs) != 2:  # len(mortality_instance):
            return logging.warning(log_message_mt)
        return func(*args, **kwargs)

    return wrapper


def check_status(func):
    def wrapper(*args, **kwargs):
        test_args = [s in args for s in all_status]
        test_kwargs = [s in kwargs.values() for s in all_status]
        if sum(test_args) + sum(test_kwargs) != 1:
            logging.warning(log_message_status)
            return
        return func(*args, **kwargs)

    return wrapper


# Probabilities

@check_status
@check_mortality_tables
def npxy(mtx, mty, x, y, n=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of years
        status: probabilities for 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods, ages or terms (udd, cfm, bal)
    Returns: probability of a group of two lives to survive at least n years
    """
    if status == 'joint-life':
        return mtx.npx(x, n, method) * mty.npx(y, n, method)
    if status == 'last-survivor':
        return 1 - mtx.nqx(x, n, method) * mty.nqx(y, n, method)

@check_status
@check_mortality_tables
def nqxy(mtx, mty, x, y, n=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of years
        status: probabilities for 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods, ages or terms
    Returns: probability of a group of two lives to extinguish before n years (udd, cfm, bal)
    """
    if status == 'joint-life':
        return 1 - mtx.npx(x, n, method) * mty.npx(y, n, method)
    if status == 'last-survivor':
        return mtx.nqx(x, n, method) * mty.nqx(y, n, method)


@check_status
@check_mortality_tables
def t_nqxy(mtx, mty, x, y, n=1, t=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of years
        status: probabilities for 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods, ages or terms
    Returns: probability of a group of two lives to survive at least n years and extinguishes
    before n+t years
    """
    return npxy(mtx, mty, x, y, t, status, method) - npxy(mtx, mty, x, y, n + t, status, method)

# Life Expectancy
@check_status
@check_mortality_tables
def exy(mtx, mty, x, y, status='joint-life', method='udd'):
    if status == 'joint-life':
        ages = np.arange(1, min(mtx.w - x, mty.w - y) + 1, 1)
    if status == 'last-survivor':
        ages = np.arange(1, max(mtx.w - x, mty.w - y) + 1, 1)
    pxy = [npxy(mtx, mty, x, y, n=age, status=status, method=method) for age in ages]
    return sum(pxy) + .5


@check_status
@check_mortality_tables
def exyn(mtx, mty, x_young, y_old, n, status='joint-life', method='udd'):
    if x_young > y_old:
        return np.nan
    dif_age = y_old - x_young

    if status == 'joint-life':
        last_age = min(mtx.w, mty.w) + 1
        ages_jl = np.linspace(0, last_age, last_age + 1)
        px = [npxy(mtx, mty, x, x + dif_age, n=1, status='joint-life', method=method) for x in ages_jl]
    if status == 'last-survivor':
        last_age = max(mtx.w, mty.w) + 1
        ages_ls = np.linspace(0, last_age, last_age + 1)
        px = [npxy(mtx, mty, x, x + dif_age + dif_age, n=1, status='last-survivor', method=method) for x in ages_ls]

    px = [p for p in px if p != 0]
    px.insert(0, 0)
    mt_xy = mt.MortalityTable(data_type='p', mt=px, perc=100, last_q=1)

    return mt_xy.exn(x=x_young, n=n, method=method)


# Multiple Life Annuities

## Life Generic Annuity 2 head
@check_status
@check_mortality_tables
def annuity_xy(mtx, mty, x, x_first_payment, x_last_payment, y, i=None, g=.0, m=1, status='joint-life', method='udd'):
    """
    Computes the present value of an annuity that starts paying 1 at age x, increasing by (1+g/100) and stops
    at age x_w, paying (1+g)^{t-1}
    :param mtx: mortality table for life x
    :param mty: mortality table for life y
    :param x: age of life x
    :param y: age of life y
    :param x_first_payment: age of x at the first payment
    :param x_last_payment: age of x at the final payment
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate quoted
    :param status: joint-life or last-survivor
    :param method: the method to approximate the fractional periods
    :return: the actuarial present value
    """

    if x_first_payment < x: return np.nan
    if x_last_payment < x_first_payment == x: return np.nan
    if int(m) != m or m <= 0: return np.nan
    if x == x_first_payment == x_last_payment: return 1
    if x < 0 or y < 0: return np.nan
    i = i / 100
    g = g / 100
    d = float((1 + g) / (1 + i))

    y_first_payment = y + (x_first_payment - x)
    y_last_payment = min(y + (x_last_payment - x), mty.w + 1)
    x_first_payment = min(x_first_payment, mtx.w + 1)
    if x_first_payment > mtx.w and y_first_payment > mty.w:
        if x_first_payment == x:
            return 1.
        return .0

    number_of_payments = int(np.round((x_last_payment - x_first_payment) * m + 1, 0))
    number_of_payments_y = int(np.round((y_last_payment - y_first_payment) * m + 1, 0))
    payments_instants = np.linspace(x_first_payment - x, x_last_payment - x, number_of_payments)
    payments_instants_y = np.linspace(y_first_payment - y, y_last_payment - y, number_of_payments_y)

    py = payments_instants
    if len(payments_instants_y) < len(payments_instants):
        py = payments_instants_y

    instalments_jl = [mtx.npx(x, n=t, method=method) * mty.npx(y, n=t, method=method) * np.power(d, t)
                      for t in py]
    sum_instalments = sum(instalments_jl)

    if status == 'last-survivor':
        instalments_x = [mtx.npx(x, n=t, method=method) * np.power(d, t) for t in payments_instants]
        instalments_y = [mty.npx(y, n=t, method=method) * np.power(d, t) for t in payments_instants_y]
        sum_instalments = sum(instalments_x) + sum(instalments_y) - sum_instalments

    return sum_instalments / np.power(1 + g, x_first_payment - x) / m


# Whole Life Annuities - 2 heads

def axy(mtx, mty, x, y, i=None, g=0, m=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: frequency of payments per unit of interest rate quoted
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods or ages (udd, cfm, bal)

    Returns: Returns the actuarial present value of an immediate whole life annuity paid for a group of two lives.
    Payments are made in the end of the periods.
    For constant annuities, pays 1 per time period.
    For fractional annuities, payments of 1/m are made m times per year at the end of the periods.
    For annuities with geometric growth, the rate is g for each payment period.
    """

    if x + 1 / m > max(mtx.w, mty.w): return 0

    return annuity_xy(mtx=mtx, mty=mty, x=x, x_first_payment=x + 1 / m, x_last_payment=max(mtx.w, mty.w) + 1,
                      y=y, i=i, g=g, m=m, status=status, method=method)

def aaxy(mtx, mty, x, y, i=None, g=0, m=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: frequency of payments per unit of interest rate quoted
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods or ages (udd, cfm, bal)

    Returns: actuarial present value of a due whole life annuity paid for a group of two lives.
    Payments are made in the beginning of the periods.
    For constant annuities, pays 1 per time period.
    For fractional annuities, payments of 1/m are made m times per year at the beginning of periods.
    For annuities with geometric growth, the rate is g for each payment period.

    """
    if x > max(mtx.w, mty.w): return 1

    return annuity_xy(mtx=mtx, mty=mty, x=x, x_first_payment=x, x_last_payment=max(mtx.w, mty.w) + 1,
                      y=y, i=i, g=g, m=m, status=status, method=method)

def t_axy(mtx, mty, x, y, i=None, g=0, m=1, defer=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: frequency of payments per unit of interest rate quoted
        defer: deferment period
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: Returns the actuarial present value of a deferred whole life annuity paid for a group of two lives.
    Payments are made in the end of the periods.
    For constant annuities, pays 1 per time period.
    For fractional annuities, payments of 1/m are made m times per year at the end of the periods.
    For annuities with geometric growth, the rate is g for each payment period.
    """

    if x + 1 / m + defer > max(mtx.w, mty.w): return 0

    return annuity_xy(mtx=mtx, mty=mty, x=x, x_first_payment=x + 1 / m + defer, x_last_payment=max(mtx.w, mty.w) + 1,
                      y=y, i=i, g=g, m=m, status=status, method=method)

def t_aaxy(mtx, mty, x, y, i=None, g=0, m=1, defer=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: frequency of payments per unit of interest rate quoted
        defer: deferment period
        status: probabilities for 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: Returns the actuarial present value of a deferred whole life annuity paid for a group of two lives.
    Payments are made in the beginning of periods.
    For constant annuities, pays 1 per time period.
    For fractional annuities, payments of 1/m are made m times per year at the beginning of the periods.
    For annuities with geometric growth, the rate is g for each payment period.
    """

    if x + defer > max(mtx.w, mty.w): return 1

    return annuity_xy(mtx=mtx, mty=mty, x=x, x_first_payment=x + defer, x_last_payment=max(mtx.w, mty.w) + 1,
                      y=y, i=i, g=g, m=m, status=status, method=method)


# Temporary Life Annuites - 2 heads

def naxy(mtx, mty, x, y, n, i=None, g=0, m=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: frequency of payments per unit of interest rate quoted
        status: probabilities for 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: Returns the actuarial present value of an immediate temporary life annuity paid for a group of two lives.
    Payments are made in the end of the periods.
    For constant annuities, pays 1 per time period.
    For fractional annuities, payments of 1/m are made m times per year at the end of the periods.
    For annuities with geometric growth, the rate is g for each payment period.
    """

    if x + 1 / m > max(mtx.w, mty.w): return 0

    return annuity_xy(mtx=mtx, mty=mty, x=x, x_first_payment=x + 1 / m, x_last_payment=x + n, y=y, i=i, g=g, m=m,
                      status=status, method=method)

def naaxy(mtx, mty, x, y, n, i=None, g=0, m=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortaity table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: frequency of payments per unit of interest rate quoted
        status: probabilities for 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: Returns the actuarial present value of a temporary life annuity due paid for a group of two lives.
    Payments are made in the beginning of periods.
    For constant annuities, pays 1 per time period.
    For fractional annuities, payments of 1/m are made m times per year at the end of the periods.
    For annuities with geometric growth, the rate is g for each payment period.
    """

    if x > max(mtx.w, mty.w): return 1

    return annuity_xy(mtx=mtx, mty=mty, x=x, x_first_payment=x, x_last_payment=x + n - 1 / m, y=y, i=i, g=g, m=m,
                      status=status, method=method)


def t_naxy(mtx, mty, x, y, n, i=None, g=0, m=1, defer=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: frequency of payments per unit of interest rate quoted
        defer: deferment period
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: Returns the actuarial present value of a deferred temporary life annuity paid for a group of two lives.
    Payments are made at the end of the periods.
    For constant annuities, pays 1 per time period.
    For fractional annuities, payments of 1/m are made m times per year at the end of the periods.
    For annuities with geometric growth, the rate is g for each payment period.
    """

    if x + 1 / m + defer > max(mtx.w, mty.w): return 0

    return annuity_xy(mtx=mtx, mty=mty, x=x, x_first_payment=x + 1 / m + defer, x_last_payment=x + n + defer, y=y,
                      i=i, g=g, m=m, status=status, method=method)


def t_naaxy(mtx, mty, x, y, n, i=None, g=0, m=1, defer=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: frequency of payments per unit of interest rate quoted
        defer: deferment period
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: Returns the actuarial present value of a deferred temporary life annuity paid for a group of two lives.
    Payments are made in the beginning of the periods.
    For constant annuities, pays 1 per time period.
    For fractional annuities, payments of 1/m are made m times per year at the end of the periods.
    For annuities with geometric growth, the rate is g for each payment period.
    """

    if x + defer > max(mtx.w, mty.w): return 1

    return annuity_xy(mtx=mtx, mty=mty, x=x, x_first_payment=x + defer, x_last_payment=x + n + defer - 1 / m, y=y,
                      i=i, g=g, m=m, status=status, method=method)

# Actuarial Present Value / Endowment Insurance - 2 heads
@check_status
@check_mortality_tables
def nExy(mtx, mty, x, y, i=None, n=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        n: number of years until term
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate due to fractional ages (udd, cfm, bal) 

    Returns: Returns the actuarial expected present value of an unitary capital paid in n years 
    term (Pure Endowment) paid upon the survival of a group of two lives.
    """

    if status == 'joint-life':
        return mtx.npx(x=x, n=n, method=method) * mty.npx(x=y, n=n, method=method) / \
            (1 + i / 100) ** n
    if status == 'last-survivor':
        return (1 - mtx.nqx(x=x, n=n, method=method) * mty.nqx(x=y, n=n, method=method)) / \
            (1 + i / 100) ** n


# Life Benefits - 2 heads

## Generic Term Life Insurance 2 head
def A_xy(mtx, mty, x, y, n, i=None, g=0, m=1, defer=0, status='joint-life', method='udd'):
    """

    Args:
        mtx: mortality table for x
        mty: mortality table for y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: frequency of payments per unit of interest rate quoted
        defer: deferment period
        status: probabilities for 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods, ages or terms

    Returns: Returns the actuarial expected present value
    """
    if defer < 0:
        return np.nan
    v = (1 + g / 100) / (1 + i / 100)
    d_m = (1 - v ** (1 / m)) * m
    ann = annuity_xy(mtx=mtx, mty=mty, x=x + defer, x_first_payment=x + defer,
                     x_last_payment=x + defer + n - 1, y=y + defer, i=i, g=g, m=m, status=status, method=method)
    pure_endow_1 = nExy(mtx, mty, x, y, i, defer, status, method)
    pure_endow_2 = nExy(mtx, mty, x + defer, y + defer, i, n, status, method)
    endow = (1 - d_m * ann)
    return pure_endow_1 * (endow - pure_endow_2)


def Axy(mtx, mty, x, y, i=None, g=0, m=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: number of periods per year at the end of which the capital is payable in case of insured event
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: returns the expected present value of a whole life insurance (i.e. net single premium),
    that pays 1 at the end of the period of death of a group of two lives.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    n = max(mtx.w, mty.w)
    return A_xy(mtx, mty, x, y, n, i, g, m, defer=0, status=status, method=method)


def Axy_(mtx, mty, x, y, i=None, g=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: returns the expected present value of a whole life insurance (i.e. net single premium), that pays 1 at the
    middle of the year of the death of a group of two lives.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    return Axy(mtx, mty, x, y, i=i, g=g, m=1, status=status, method=method) * (1 + i / 100) ** .5


def t_Axy(mtx, mty, x, y, i=None, g=0, m=1, defer=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: number of periods per year at the end of which the capital is payable in case of insured event
        defer: deferment period
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: returns the expected present value of a deferred whole life insurance (i.e. net single premium),
    that pays 1 at the end of the period of death of a group of two lives.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    n = max(mtx.w, mty.w)
    return A_xy(mtx, mty, x, y, n, i=i, g=g, m=m, defer=defer, status=status, method=method)


def t_Axy_(mtx, mty, x, y, i=None, g=0, defer=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        defer: deferment period
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: returns the expected present value of a deferred whole life insurance
    (i.e. net single premium), that pays 1 at the middle of the year of the death of a group of two lives.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    return t_Axy(mtx, mty, x, y, i=i, g=g, m=1, defer=defer, status=status, method=method) * (1 + i / 100) ** .5

# Term Life Insurance - 2 heads

def nAxy(mtx, mty, x, y, n, i=None, g=0, m=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: number of periods per year at the end of which the capital is payable in case of insured event
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: returns the expected present value of a temporary life insurance (i.e. net single premium),
    that pays 1 at the end of the period of death of a group of two lives.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    return A_xy(mtx, mty, x, y, n, i, g=g, m=m, defer=0, status=status, method=method)


def nAxy_(mtx, mty, x, y, n, i=None, g=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        status: probabilities for 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: returns the expected present value of a temporary life insurance (i.e. net single premium),
    that pays 1 at the middle of the year of the death of a group of two lives.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    return A_xy(mtx, mty, x, y, n, i, g, m=1, defer=0, status=status, method=method) * (1 + i / 100) ** .5


def t_nAxy(mtx, mty, x, y, n, i=None, g=0, m=1, defer=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: number of periods per year at the end of which the capital is payable in case of insured event
        defer: deferment period
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods, ages or terms

    Returns: returns the expected present value of a deferred temporary life insurance (i.e. net single premium),
    that pays 1 at the end of the period of death of a group of two lives.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    return A_xy(mtx, mty, x, y, n, i, g, m, defer=defer, status=status, method=method)


def t_nAxy_(mtx, mty, x, y, n, i=None, g=0, defer=0, status='joint-life', method='udd'):
    """
        Args:
            mtx: mortality table for life x
            mty: mortality table for life y
            x: age of life x
            y: age of life y
            n: number of periods
            i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
            g: growth rate (flat rate) in percentage, e.g., 2 for 2%
            defer: deferment period
            status: 'joint-life' or 'last-survivor' status
            method: the method to approximate the fractional periods and ages

        Returns: returns the expected present value of a deferred temporary life insurance (i.e. net single premium),
        that pays 1 in the middle of the year of death of a group of two lives.
        For contracts with geometric growth, the rate is g for each payment period.
        """
    return A_xy(mtx, mty, x, y, n, i, g, m=1, defer=defer, status=status, method=method) * (1 + i / 100) ** .5

# Endowment Life Insurance - 2 heads
def nAExy(mtx, mty, x, y, n, i=None, g=0, m=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: number of periods per year at the end of which the capital is payable in case of insured event
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods, ages or terms

    Returns: returns the expected present value of an endowment life insurance (i.e. net single premium),
    that pays 1 at the end of the period of death of a group of two lives or pays 1 if the group is alive at the end
    of the contract.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    return nAxy(mtx, mty, x, y, n, i, g=g, m=m, status=status, method=method) + \
        nExy(mtx, mty, x, y, i, n=n, status=status, method=method)


def nAExy_(mtx, mty, x, y, n, i=None, g=0, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: returns the expected present value of an endowment life insurance (i.e. net single premium),
    that pays 1 in the middle of the period of death of a group of two lives or pays 1 if the group is alive at the end
    of the contract.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    return nAxy_(mtx, mty, x, y, n, i, g=g, status=status, method=method) + \
        nExy(mtx, mty, x, y, i, n=n, status=status, method=method)


def t_nAExy(mtx, mty, x, y, n, i=None, g=.0, m=1, defer=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        m: number of periods per year at the end of which the capital is payable in case of insured event
        defer: deferment period
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: returns the expected present value of a deferred endowment life insurance (i.e. net single premium),
    that pays 1 at the end of the period of death of a group of two lives or pays 1 if the group is alive at the end
    of the contract.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    return t_nAxy(mtx, mty, x, y, n, i, g=g, m=m, defer=defer, status=status, method=method) + \
        nExy(mtx, mty, x, y, i, n=defer + n, status=status, method=method)


def t_nAExy_(mtx, mty, x, y, n, i=None, g=.0, defer=1, status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        g: growth rate (flat rate) in percentage, e.g., 2 for 2%
        defer: deferment period
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: returns the expected present value of a deferred endowment life insurance (i.e. net single premium),
    that pays 1 at the moment of death of a group of two lives or pays 1 if the group is alive at the end
    of the contract.
    For contracts with geometric growth, the rate is g for each payment period.
    """
    return t_nAxy_(mtx, mty, x, y, n, i, g=g, defer=defer, status=status, method=method) + \
        nExy(mtx, mty, x, y, i, n=defer + n, status=status, method=method)




#################################################
@check_status
@check_mortality_tables
def _create_joint_mortality_table(mt_old, mt_young, age_dif=0, status='joint-life'):
    return mt2h.MortalityTable2Heads(mt_old=mt_old, mt_young=mt_young, age_dif=age_dif, status=status)


@check_status
@check_mortality_tables
def t_nIArxy(mtx, mty, x, y, n, i=None, defer=0, first_payment=1, inc=1,
             status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        defer: deferment period
        first_payment: amount of capital in the first year
        inc: rate of the progression (in monetary units)
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: Returns the Expected Present Value of a deferred Term Life Insurance that pays
    1+k, at the end of year of death of a group of two individuals, if death occurs between ages
    x+t+k and x+t+k+1, for k=0, 1,..., n-1.
    The capital of the first year may differ from the rate of the progression.
    """
    if n < 0 or defer < 0 or x < 0 or y < 0: return np.nan
    if first_payment + n * inc < 0: return np.nan

    payment1 = t_nAxy(mtx, mty, x, y, n, i, g=.0, m=1, defer=defer, status=status, method=method)
    payment1 *= first_payment

    payment2 = [t_nAxy(mtx, mty, x, y, n - j, i, g=.0, m=1, defer=defer + j, status=status, method=method)
                for j in range(1, min(max(mtx.w - x, mty.w - y), n))]
    payment2 = sum(payment2)
    payment2 *= inc
    payment_jl = payment1 + payment2

    if status == 'joint-life':
        return payment_jl

    payment_x = mi.t_nIArx(mtx, x, n, defer, i, first_payment, inc, method)
    payment_y = mi.t_nIArx(mty, y, n, defer, i, first_payment, inc, method)

    return payment_x+payment_y-payment_jl

@check_status
@check_mortality_tables
def t_nIArxy_(mtx, mty, x, y, n, i=None, defer=0, first_payment=1, inc=1,
             status='joint-life', method='udd'):
    """
    Args:
        mtx: mortality table for life x
        mty: mortality table for life y
        x: age of life x
        y: age of life y
        n: number of periods
        i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
        defer: deferment period
        first_payment: amount of capital in the first year
        inc: rate of the progression (in monetary units)
        status: 'joint-life' or 'last-survivor' status
        method: the method to approximate the fractional periods and ages (udd, cfm, bal)

    Returns: Returns the Expected Present Value of a deferred Term Life Insurance that pays
    1+k, at the moment of death of a group of two individuals, if death occurs between ages
    x+t+k and x+t+k+1, for k=0, 1,..., n-1.
    The capital of the first year may differ from the rate of the progression.
    """

    return t_nIArxy(mtx, mty, x, y, n, i, defer, first_payment, inc, status, method) * (1 + i / 100) ** .5


__author__ = "PedroCR"

import numpy as np
import pandas as pd
from lifeActuary.mortality_table import MortalityTable


class CommutationFunctions(MortalityTable):
    """
    Instantiates, for a specific mortality table and interest rate, all the usual commutation functions:
    Dx, Nx, Sx, Cx, Mx, Rx.

    :param i: interest rate, in percentage. Use 5 for 5% :param g: rate of growing (in percentage), for capitals
    evolving geometrically :param data_type: Use 'l' for lx, 'p' for px and 'q' for qx.
    param mt: the mortality.
    table, in array format, according to the data_type defined.
    param perc: The percentage of qx to use, e.g., use 50 for 50%.
    param app_cont: Use 'True' for continuous approach (deaths occur, in average, in the middle of
    the year and payments are due in the moment of death) or 'False' for considering that death payments are due in
    the end of the year

    :return: the commutation symbols Dx, Nx, Sx, Cx, Mx, Rx.
    """

    def __init__(self, i=None, g=0, data_type='q', mt=None, perc=100, app_cont=False):
        MortalityTable.__init__(self, data_type, mt, perc)
        if i is None:
            return
        self.__i = i / 100.
        self.__g = g / 100.
        self.__v = 1 / (1 + self.__i)
        self.__d = (1 + self.__g) / (1 + self.__i)
        self.__app_cont = app_cont
        self.__cont = np.sqrt(1 + self.__i)

        # self.__Dx = np.array([self.lx[x] * np.power(self__d, x) for x in range(len(self.lx))])
        self.__Dx = self.lx[:-1] * np.power(self.__d, range(len(self.lx[:-1])))
        self.__Nx = np.array([np.sum(self.__Dx[x:]) for x in range(len(self.lx[:-1]))])
        self.__Sx = np.array([np.sum(self.__Nx[x:]) for x in range(len(self.__Nx))])
        self.__Cx = self.dx * np.power(self.__d, range(1, len(self.dx) + 1))
        self.__Mx = np.array([np.sum(self.__Cx[x:]) for x in range(len(self.__Cx))])
        self.__Rx = np.array([np.sum(self.__Mx[x:]) for x in range(len(self.__Mx))])
        if self.__app_cont:
            self.__Cx = self.__Cx * self.__cont
            self.__Mx = self.__Mx * self.__cont
            self.__Rx = self.__Rx * self.__cont

    def __repr__(self):
        return f"{self.__class__.__name__}{self.i, self.g, self.data_type, self.mt, self.perc, self.app_cont}"

    # getters and setters
    @property
    def i(self):
        return self.__i * 100

    @property
    def g(self):
        return self.__g * 100

    @property
    def v(self):
        return self.__v

    @property
    def d(self):
        return self.__d

    @property
    def app_cont(self):
        return self.__app_cont

    @property
    def cont(self):
        return self.__cont

    @property
    def Dx(self):
        return self.__Dx

    @property
    def Nx(self):
        return self.__Nx

    @property
    def Sx(self):
        return self.__Sx

    @property
    def Cx(self):
        return self.__Cx

    @property
    def Mx(self):
        return self.__Mx

    @property
    def Rx(self):
        return self.__Rx

    def df_commutation_table(self):
        data = {'Dx': self.__Dx, 'Nx': self.__Nx, 'Sx': self.__Sx, 'Cx': self.__Cx, 'Mx': self.__Mx, 'Rx': self.__Rx}
        df = pd.DataFrame(data)
        data_lf = self.df_life_table()
        df = pd.concat([data_lf, df], axis=1, sort=False)
        return df

    ### Life Annuities

    ## Whole Life Annuities

    def ax(self, x, m=1):
        """
        Returns the actuarial present value of an immediate whole life annuity of 1 per year
        Payments of 1/m are made m times per year at the end of the periods.

        :param x: age at the beginning of the contract
        :param m: number of payments per year

        :return: Expected Present Value (EPV) for payments of 1/m
        """
        if x < 0:
            return np.nan
        if m < 0:
            return np.nan
        if x >= self.w:
            return 0
        aux = self.__Nx[x + 1] / self.__Dx[x] / (1 + self.__g) + (m - 1) / (m * 2)
        self.msn.append(f"ax_{x}={self.__Nx[x + 1]}/{self.__Dx[x]}+({m}-1)/({m}*2)")
        return aux

    def aax(self, x, m=1):
        """
        Returns the actuarial present value of a due whole annuity of 1 per year.
        Payments of 1/m are made m times per year at the beginning of the periods.

        :param x: age at the beginning of the contract
        :param m: number of payments per year (used to quote the interest rate)

        :return: Expected Present Value (EPV) for payments of 1/m
        """
        if x > self.w:
            return 1
        aux = self.__Nx[x] / self.__Dx[x] - (m - 1) / (m * 2)
        self.msn.append(f"aax_{x}={self.__Nx[x]}/{self.__Dx[x]}-({m}-1)/({m}*2)")
        return aux

    # Deferred Whole Life Annuities

    def t_ax(self, x, m=1, defer=0):
        """
        Returns the actuarial present value of an (immediate) annuity of 1 per year
        (whole life annuity-late), deferred t years. Payable 'm' times per year at the end of the period. Payable 'm'
        times per year at the end of the period

        :param x: age at the beginning of the contract
        :param m: number of payments per year used to quote the interest rate
        :param defer: deferment period

        :return: Expected Present Value (EPV) for payments of 1/m
        """
        # note: nEx discounts the growth rate np.power(1 + self.__g, defer + 1)
        aux = self.ax(x + defer, m) * self.nEx(x, defer)
        if aux > 0:
            self.msn.append(f"{defer}_ax_{x}=[{self.__Nx[x + 1 + defer]}/{self.__Dx[x + defer]}+({m} + 1)/({m}*2)]"
                            f"*{self.__Dx[x + defer]}/{self.__Dx[x]}")
        return aux

    def t_aax(self, x, m=1, defer=0):
        """
        Returns the actuarial present value of an (immediate) annuity of 1 per year
        (whole life annuity-anticipatory), deferred t years. Payable 'm' times per year at the beginning of the period

        :param x: age at the beginning of the contract
        :param m: number of payments per period used to quote the interest rate
        :param defer: deferment period

        :return: Expected Present Value (EPV) for payments of 1/m
        """
        aux = self.aax(x + defer, m) * self.nEx(x, defer)
        if x + defer < self.w:
            self.msn.append(f"{defer}_aax_{x}=[{self.__Nx[x + defer]}/{self.__Dx[x + defer]}-({m}-1)/({m}*2)]"
                            f"*{self.__Dx[x + defer]}/{self.__Dx[x]}")
        return aux

    ## Temporary Life Annuities

    def nax(self, x, n, m=1):
        """
        Returns the actuarial present value of an immediate temporary life annuity: n-year temporary
        life annuity-late. Payable 'm' times per year at the ends of the period

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param m: number of payments per year

        :return: Expected Present Value (EPV) for payments of 1/m
        """
        if x >= self.w:
            return 0
        if x < 0:
            return np.nan
        if m < 0:
            return np.nan
        if n < 0:
            return 0

        if x + 1 + n <= self.w:
            aux = (self.__Nx[x + 1] - self.__Nx[x + 1 + n]) / self.__Dx[x] / (1 + self.__g) + \
                  (m - 1) / (m * 2) * (1 - self.nEx(x, n))
            self.msn.append(f"{n}_ax_{x}={self.__Nx[x + 1] - self.__Nx[x + 1 + n]}/{self.__Dx[x]}+({m}-1)/({m}*2)*"
                            f"(1-{self.__Dx[x + n]}/{self.__Dx[x]})")
        else:
            return self.ax(x=x, m=m)

        return aux

    def naax(self, x, n, m=1):
        """
        Returns the actuarial present value of a due temporary life annuity: n-year temporary
        life annuity-anticipatory. Payable 'm' times per year at the beginning of the period

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param m: number of payments per year

        :return: Expected Present Value (EPV) for payments of 1/m
        """
        if x >= self.w or n == 1:
            return 1
        if x < 0:
            return np.nan
        if m < 0:
            return np.nan
        if n < 0:
            return 0

        if x + 1 + n <= self.w + 1:
            aux = (self.__Nx[x] - self.__Nx[x + n]) / self.__Dx[x] - (m - 1) / (m * 2) * (1 - self.nEx(x, n))
            if x + 1 + n <= self.w:
                Nx2 = self.__Nx[x + 1 + n]
            else:
                Nx2 = 0
            self.msn.append(
                f"{n}_aax_{x}={self.__Nx[x + 1] - Nx2}/{self.__Dx[x]}*(1+{self.__g}) + ({m}+1)/({m}*2)*"
                f"(1-{self.__Dx[x + n]}/{self.__Dx[x]})")
        else:
            return self.aax(x=x, m=m)
        return aux

    # Deferred Temporary Life Annuities

    def t_nax(self, x, n, m=1, defer=0):
        """
        Returns the actuarial present value of a deferred temporary life annuity: n-year temporary
        life annuity-late, deferred t periods. Payable 'm' per year at the ends of the period

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param m: number of payments per year
        :param defer: deferment period (number of years)

        :return: Expected Present Value (EPV) for payments of 1/m
        """
        aux = self.nax(x + defer, n, m) * self.nEx(x, defer)
        if x + 1 + n + defer <= self.w:
            self.msn.append(
                f"{defer}|{n}_ax_{x}=[{self.__Nx[x + 1 + defer] - self.__Nx[x + 1 + n + defer]}/{self.__Dx[x + defer]}"
                f"+ ({m}-1)/({m}*2)*(1-{self.__Dx[x + n + defer]}/{self.__Dx[x + defer]})]"
                f"*{self.__Dx[x + defer]}/{self.__Dx[x]}")
        else:
            return self.t_ax(x=x, m=m, defer=defer)
        return aux

    def t_naax(self, x, n, m=1, defer=0):
        """
        Returns the actuarial present value of a due temporary life annuity: n-year temporary
        life annuity-anticipatory, deferred t periods. Payable 'm' times per year at the beginning of the period

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param m: number of payments per year
        :param defer: deferment period (number of years)

        :return: Expected Present Value (EPV) for payments of 1/m
        """
        aux = self.naax(x + defer, n, m) * self.nEx(x, defer)
        if x + 1 + n + defer <= self.w + 1:
            if x + 1 + n + defer <= self.w:
                Nx2 = self.__Nx[x + 1 + n + defer]
            else:
                Nx2 = 0
            self.msn.append(
                f"{defer}|{n}_aax_{x}=[{self.__Nx[x + 1 + defer] - Nx2}/{self.__Dx[x + defer]}"
                f"+({m}+1)/({m}*2)*(1-{self.__Dx[x + n + defer]}/{self.__Dx[x + defer]})]"
                f"*{self.__Dx[x + defer]}/{self.__Dx[x]}")
        else:
            return self.t_aax(x=x, m=m, defer=defer)
        return aux

    ## Life Annuities with variable terms

    # Annuities Increasing and Decreasing Arithmetically

    def t_nIax(self, x, n, m=1, defer=0, first_amount=1, increase_amount=1):
        """
        Returns the actuarial present value of an immediate n term life annuity, deferred $t$ periods,
        with payments evolving in arithmetic progression. Payments of 1/m are made m times per year at the end of the periods.
        First amount and Increase amount may be different.
        For decreasing life annuities, the Increase Amount should be negative.

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param m: number of payments per year
        :param defer: umber of deferment years
        :param first_amount: amount of the first payment
        :param increase_amount: amount of the increase amount

        :return: Expected Present Value (EPV) for payments of 1/m
        """

        if first_amount + (n - 1) * increase_amount < 0:
            return np.nan
        if x + n + defer > self.w:
            return .0

        term1 = first_amount * self.t_nax(x=x, n=n, m=m, defer=defer)
        list_increases = [increase_amount * self.t_nax(x=x + defer, n=n - j, m=m, defer=defer + j)
                          for j in range(1, n)]

        return term1 + sum(list_increases)

    def t_nIaax(self, x, n, m=1, defer=0, first_amount=1, increase_amount=1):
        """
        Returns the actuarial present value of an immediate n term life annuity, deferred t periods,
        with payments evolving in arithmetic progression. Payments of 1/m are made m times per year at the beginning
        of the periods.
        First amount and Increase amount may be different.
        For decreasing life annuities, the Increase Amount should be negative.

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param m: number of payments per year
        :param defer: umber of deferment years
        :param first_amount: amount of the first payment
        :param increase_amount: amount of the increase amount

        :return: Expected Present Value (EPV) for payments of 1/m
        """

        if first_amount + (n - 1) * increase_amount < 0:
            return np.nan
        if x + n + defer > self.w:
            return .0

        term1 = first_amount * self.t_naax(x=x, n=n, m=m, defer=defer)
        list_increases = [increase_amount * self.t_nax(x=x + defer, n=n - j, m=m, defer=defer + j - 1)
                          for j in range(1, n)]

        return term1 + sum(list_increases)

    # Present Value of a series of cash-flows
    def present_value(self, probs, age, spot_rates, capital):
        """
        Computes the expected present value of a cash-flows, that can be contingent on some probabilities.
        Payments are considered at the end of the period.

        :param probs: vector of probabilities. For using the instantiated actuarial table, introduce probs=None
        :param age: age at the beginning of the contract
        :param spot_rates: vector of interest rates for the considered time periods
        :param capital: vector of cash-flow amounts

        :return: the expected present value of a cash-flow, that can be contingent on some probabilities.
        """
        if len(spot_rates) != len(capital):
            return np.nan
        probs_ = None
        if probs is None:
            if age is None:
                return np.nan
            else:
                probs_ = [self.npx(age, n + 1) for n in range(len(capital))]

        if isinstance(probs, list):
            if len(probs) == len(spot_rates):
                probs_ = probs
            else:
                return np.nan

        if isinstance(probs, (float, int)):
            probs_ = [probs] * len(capital)
        discount = 1 + np.array(spot_rates) / 100.
        discount = np.cumprod(1 / discount)

        return sum([p * capital[idx_p] * discount[idx_p] for idx_p, p in enumerate(probs_)])

    ### Life Insurances

    ## Pure Endowment / Actuarial Present Value

    def nEx(self, x, n):
        """
        Actuarial Present Value of a Pure endowment or Deferred capital

        :param x: age at the beginning of the contract
        :param n: years until payment, if x is alive

        :return: actuarial present value of a pure endowment of 1 paid at age x+n
        """
        if x < 0:
            return np.nan
        if n <= 0:
            return 1
        if x + n > self.w:
            return 0.
        D_x = self.__Dx[x]
        D_x_n = self.__Dx[x + n]
        self.msn.append(f"{n}_E_{x}={D_x_n} / {D_x}")
        # note: nEx discounts the growth rate np.power(1 + self.__g, defer + 1) so only survival is considered
        return D_x_n / D_x / np.power(1 + self.__g, n)

    ## Whole Life Insurance

    def Ax(self, x):
        """
        Expected Present Value (EPV) of a Whole life insurance that pays 1 at the end of the year of death.

        :param x: age at the beginning of the contract

        :return: net single premium of a whole life insurance, that pays 1, at the
        end of the year of death.
        """
        if x < 0:
            return np.nan
        if x > self.w:
            return self.__v  # it will die before year's end, because already attained age>w
        D_x = self.__Dx[x]
        if self.__app_cont:
            M_x = self.__Mx[x] / self.__cont
        else:
            M_x = self.__Mx[x]
        self.msn.append(f"A_{x}={M_x} / {D_x}")
        return M_x / D_x / (1 + self.__g)

    def Ax_(self, x):
        """
        Expected Present Value (EPV) of a Whole life insurance that pays 1 at the moment of death.

        :param x: age at the beginning of the contract

        :return: net single premium of a whole life insurance, that pays 1, at the moment of death.
        """
        if x < 0:
            return np.nan
        if x > self.w:  # it will die before year's end, because already attained age>w
            return self.__v ** .5
        D_x = self.__Dx[x]
        if self.__app_cont:
            M_x = self.__Mx[x]
        else:
            M_x = self.__Mx[x] * self.__cont
        self.msn.append(f"A_{x}_={M_x} / {D_x}")
        return M_x / D_x / (1 + self.__g)

    # Deferred Whole Life Insurances
    def t_Ax(self, x, defer=0):
        """
        Expected Present Value (EPV) of a Deferred Whole life insurance, that pays 1 in the end of the year of death

        :param x: age at the beginning of the contract
        :param defer: deferment period
        :return: net single premium of a deferred whole life insurance that pays 1, at the
        end of the year of death.
        """
        aux = self.nEx(x, defer) * self.Ax(x + defer)
        self.msn.append(f"{defer}|_A_{x}={defer}_E_{x}*A_{x + defer}")
        return aux

    def t_Ax_(self, x, defer=0):
        """
        Expected Present Value (EPV) of a Deferred Whole life insurance, that pays 1 in the moment of death

        :param x: age at the beginning of the contract
        :param defer: deferment period
        :return: net single premium of a deferred whole life insurance that pays 1, at the moment of death.
        """
        aux = self.nEx(x, defer) * self.Ax_(x + defer)
        self.msn.append(f"{defer}|_A_{x}_={defer}_E_{x}*A_{x + defer}_")
        return aux

    ## Term Life Insurance

    def nAx(self, x, n):
        """
        Expected Present Value (EPV) of a Term life insurance that pays 1 in the end of the year of death, if x dies before age x+n

        :param x: age at the beginning of the contract
        :param n: period of the contract

        :return: net single premium of a Term Life Insurance that pays 1 in the end of the year of death
        """
        if x < 0:
            return np.nan
        if n < 0:
            return np.nan
        if x + n > self.w:
            return self.Ax(x)
        D_x = self.__Dx[x]
        if self.__app_cont:
            M_x = self.__Mx[x] / self.__cont
            M_x_n = self.__Mx[x + n] / self.__cont
        else:
            M_x = self.__Mx[x]
            M_x_n = self.__Mx[x + n]
        self.msn.append(f"{n}_A_{x}=({M_x}-{M_x_n}) / {D_x}")
        return (M_x - M_x_n) / D_x / (1 + self.__g)

    def nAx_(self, x, n):
        """
        Expected Present Value (EPV) of a Term life insurance that pays 1 in the moment of death, if x dies before age x+n

        :param x: age at the beginning of the contract
        :param n: period of the contract

        :return: net single premium of a Term Life Insurance that pays 1 in the moment of death
        """
        if x < 0:
            return np.nan
        if n < 0:
            return np.nan
        if x + n > self.w:
            return self.Ax(x) * self.__cont
        D_x = self.__Dx[x]
        if self.__app_cont:
            M_x = self.__Mx[x]
            M_x_n = self.__Mx[x + n]
        else:
            M_x = self.__Mx[x] * self.__cont
            M_x_n = self.__Mx[x + n] * self.__cont
        self.msn.append(f"{n}_A_{x}_=({M_x}-{M_x_n}) / {D_x}")
        return (M_x - M_x_n) / D_x / (1 + self.__g)

    # Deferred Term Life Insurances

    def t_nAx(self, x, n, defer=0):
        """
        Expected Present Value (EPV) of a t-years Deferred Term life insurance that pays 1 in the end of the year of death,
        if x dies between ages x+t and x+t+n

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param defer: deferment period (number of years)

        :return: net single premium of a Deferred Term Life Insurance that pays 1 in the end of the year of death
        """
        aux = self.nEx(x, defer) * self.nAx(x + defer, n)
        self.msn.append(f"{defer}|{n}_A_{x}={defer}_E_{x}*{n}_A_{x + defer}")
        return aux

    def t_nAx_(self, x, n, defer=0):
        """
        Expected Present Value (EPV) of a t-years Deferred Term life insurance that pays 1 in the moment of death,
        if x dies between ages x+t and x+t+n

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param defer: deferment period (number of years)

        :return: net single premium of a Deferred Term Life Insurance that pays 1 in the moment of death
        """
        aux = self.nEx(x, defer) * self.nAx_(x + defer, n)
        self.msn.append(f"{defer}|{n}_A_{x}_={defer}_E_{x}*{n}_A_{x + defer}_")
        return aux

    ## Endowment Insurance

    def nAEx(self, x, n):
        """
        Expected Present Value (EPV) of an Endowment Insurance that pays 1 in the end of the year of death, if x dies between
        x and x+n or pays 1 if x is alive at age x+n

        :param x: age at the beginning of the contract
        :param n: period of the contract

        :return: net single premium of an Endowment Insurance. Death coverage is paid at the end of the year of death
        """
        self.msn.append(f"{n}_AE_{x}={n}_A_{x}+{n}_E_{x}")
        return self.nAx(x, n) + self.nEx(x, n)

    def nAEx_(self, x, n):
        """
        Expected Present Value (EPV) of an Endowment Insurance that pays 1 in the moment of death, if x dies between
        x and x+n or pays 1 if x is alive at age x+n

        :param x: age at the beginning of the contract
        :param n: period of the contract

        :return: net single premium of an Endowment Insurance. Death coverage is paid at the moment of death
        """
        aux = self.nAx_(x, n) + self.nEx(x, n)
        self.msn.append(f"{n}_AE_{x}_={n}_A_{x}_+{n}_E_{x}")
        return aux

    # Deferred Endowment Insurance

    def t_nAEx(self, x, n, defer=0):
        """
        Expected Present Value (EPV) of t-years Deferred Endowment Insurance that pays 1 in the end of the year of death,
        if x dies between ages x+t and x+t+n or pays 1 if x is alive at age x+t+n

        :param x: age at the beginning of the contract
        :param n: period of the contract
        :param defer: deferment period (number of years)

        :return: net single premium of a Deferred Endowment Insurance. Death coverage is paid at the end of the year of death
        """
        aux = self.nEx(x, defer) * self.nAEx(x + defer, n)
        self.msn.append(f"{defer}|{n}_AE_{x}={defer}_E_{x}*{n}_AE_{x + defer}")
        return aux

    def t_nAEx_(self, x, n, defer=0):
        """
        Expected Present Value (EPV) of t-years Deferred Endowment Insurance that pays 1 in the moment of death,
        if x dies between ages x+t and x+t+n or pays 1 if x is alive at age x+t+n

        :param x: age at the beginning of the contract
        :param n: period of the contract
        :param defer: deferment period (number of years)

        :return: net single premium of a Deferred Endowment Insurance. Death coverage is paid at the moment of death
        """
        aux = self.nEx(x, defer) * self.nAEx_(x + defer, n)
        self.msn.append(f"{defer}|{n}_AE_{x}={defer}_E_{x}*{n}_AE_{x + defer}_")
        return aux

    ## Term Life Insurance with Variable Capitals

    def IAx(self, x):
        """
        Expected Present Value (EPV) of Term Life Insurance that pays 1+k, at the end of year of death, if death occurs
        between ages x+k and x+k+1, for k=0,1,...

        :param x: age at the beginning of the contract

        :return: net single premium of a Term Life Insurance with capitals increasing arithmetically, with first capital
        equal to the rate of progression (the increase amount). The payment is made at the end of the year of death.
        """
        if x < 0:
            return np.nan
        if x > self.w:
            return self.__v  # it will die before year's end, because already attained age>w
        D_x = self.__Dx[x]
        if self.__app_cont:
            R_x = self.__Rx[x] / self.__cont
        else:
            R_x = self.__Rx[x]
        self.msn.append(f"A_{x}={R_x} / {D_x}")
        return R_x / D_x

    def IAx_(self, x):
        """
        Expected Present Value (EPV) of Term Life Insurance that pays 1+k, at the moment of death, if death occurs
        between ages x+k and x+k+1, for k=0,1,...

        :param x: age at the beginning of the contract

        :return: net single premium of a Term Life Insurance with capitals increasing arithmetically, with first capital
        equal to the rate of progression (the increase amount). The payment is made at the moment of death.
        """
        if x < 0:
            return np.nan
        if x > self.w:
            return self.__v ** 0.5  # it will die before year's end, because already attained age>w
        D_x = self.__Dx[x]
        if self.__app_cont:
            R_x = self.__Rx[x]
        else:
            R_x = self.__Rx[x] * self.__cont
        self.msn.append(f"A_{x}={R_x} / {D_x}")
        return R_x / D_x

    def nIAx(self, x, n):
        """
        Expected Present Value (EPV) of Term Life Insurance that pays 1+k, at the end of year of death, if death occurs
        between ages x+k and x+k+1, for k=0,1,...,n-1.

        :param x: age at the beginning of the contract

        :return: net single premium of a Term Life Insurance with capitals increasing arithmetically, with first capital
        equal to the rate of progression (the increase amount). The payment is made at the end of the year of death.
        """
        if x < 0:
            return np.nan
        if n < 0:
            return np.nan
        if x > self.w:
            return self.__v  # it will die before year's end, because already attained age>w
        D_x = self.__Dx[x]
        if self.__app_cont:
            M_x_n = self.__Mx[x + n] / self.__cont
            R_x = self.__Rx[x] / self.__cont
            R_x_n = self.__Rx[x + n] / self.__cont
        else:
            M_x_n = self.__Mx[x + n]
            R_x = self.__Rx[x]
            R_x_n = self.__Rx[x + n]
        self.msn.append(f"A_{x}=({R_x}-{R_x_n}-{n}x{M_x_n} / {D_x}")
        return (R_x - R_x_n - n * M_x_n) / D_x

    def nIAx_(self, x, n):
        """
        Expected Present Value (EPV) of Term Life Insurance that pays 1+k, at the moment of death, if death occurs
        between ages x+k and x+k+1, for k=0,1,...n-1.

        :param x: age at the beginning of the contract

        :return: net single premium of a Term Life Insurance with capitals increasing arithmetically, with first capital
        equal to the rate of progression (the increase amount). The payment is made at the moment of death.
        """
        if x < 0:
            return np.nan
        if n < 0:
            return np.nan
        if x > self.w:
            return self.__v ** 0.5  # it will die before year's end, because already attained age>w
        D_x = self.__Dx[x]
        if self.__app_cont:
            M_x_n = self.__Mx[x + n]
            R_x = self.__Rx[x]
            R_x_n = self.__Rx[x + n]
        else:
            M_x_n = self.__Mx[x + n] * self.__cont
            R_x = self.__Rx[x] * self.__cont
            R_x_n = self.__Rx[x + n] * self.__cont
        self.msn.append(f"A_{x}=({R_x}-{R_x_n}-{n}x{M_x_n} / {D_x}")
        return (R_x - R_x_n - n * M_x_n) / D_x

    ## Variable Capitals increasing/decreasing arithmetically

    def t_nIArx(self, x, n, defer=0, first_amount=1, increase_amount=1):
        """
        Expected Present Value (EPV) of a Term Life Insurance that pays (first_amount + k*increase_amount), at the end of
        the year if death occurs between ages x+k and x+k+1, for k=0,..., n-1.
        Allows the computation of EPV for decreasing capitals.
        The first amount may differ from the increasing/decreasing amount.

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param first_amount: insured amount in the first year of the contract
        :param increase_amount: rate of increasing (if increase_amount>0) or decreasing (if increase_amount<0)

        :return: net single premium of a Term Life Insurance with capitals increasing arithmetically, with first capital
        equal to the rate of progression (the increase amount). The payment is made at the end of the year of death.
        """

        if first_amount + (n - 1) * increase_amount < 0:
            return np.nan

        term1 = first_amount * self.t_nAx(x=x, n=n, defer=defer)
        list_increases = [increase_amount * self.t_nAx(x=x, n=n - j, defer=defer + j)
                          for j in range(1, n)]

        return term1 + sum(list_increases)

    def t_nIArx_(self, x, n, defer=0, first_amount=1, increase_amount=1):
        """
        Expected Present Value (EPV) of a Term Life Insurance that pays (first_amount + k*increase_amount), at the moment of death,
        if death occurs between ages x+k and x+k+1, for k=0,..., n-1.
        Allows the computation of EPV for decreasing capitals.
        The first amount may differ from the increasing/decreasing amount.

        :param x: age at the beginning of the contract
        :param n: number of years of the contract
        :param first_amount: insured amount in the first year of the contract
        :param increase_amount: rate of increasing (if increase_amount>0) or decreasing (if increase_amount<0)

        :return: net single premium of a Term Life Insurance with capitals increasing arithmetically, with first capital
        equal to the rate of progression (the increase amount). The payment is made at the moment of death.
        """

        if first_amount + (n - 1) * increase_amount < 0:
            return np.nan

        term1 = first_amount * self.t_nAx_(x=x, n=n, defer=defer)
        list_increases = [increase_amount * self.t_nAx_(x=x, n=n - j, defer=defer + j)
                          for j in range(1, n)]

        return term1 + sum(list_increases)



__author__ = "PedroCR"

import numpy as np
from lifeActuary import annuities


def A_x(mt, x, x_first, x_last, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value (EPV) of a Life Insurance that pays 1 at the end of the year of death.
    :param mt: mortality table for life x
    :param x: age at the beginning of the contract
    :param x_first: age of first payment
    :param x_last: age of final payment
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV)
    """
    if x_first < x: return np.nan
    if x_last < x_first == x: return np.nan
    if x == x_first == x_last: return 0
    i = i / 100
    g = g / 100
    v = float((1 + g) / (1 + i))
    number_of_payments = int((x_last - x_first) + 1)
    if number_of_payments < 1: return .0  # according to the mortality table x is going to die before x+1
    payments_instants = np.linspace(x_first - x, x_last - x, number_of_payments)
    instalments = [mt.npx(x, n=t - 1, method=method) * mt.nqx(x + t - 1, n=1, method=method) * np.power(v, t)
                   for t in payments_instants]
    instalments = np.array(instalments) / np.power(1 + g, payments_instants[0])
    return np.sum(instalments)


def Ax(mt, x, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value (EPV) of a Whole Life Insurance that pays 1 at the end of the year of death.
    :param mt: mortality table for life x
    :param x: age at the beginning of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Whole Life Insurance (end of the year of death)
    """

    if x > mt.w:
        return np.power(1 + i / 100, -1)  # it will die before year's end, because already attained age>w
    return A_x(mt=mt, x=x, x_first=x + 1, x_last=mt.w + 1, i=i, g=g, method=method)


def Ax_(mt, x, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value (EPV) of a Whole Life Insurance that pays 1 at the moment of death.
    :param mt: mortality table for life x
    :param x: age at the beginning of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Whole Life Insurance (moment of death)
    """
    if x > mt.w:
        return np.power(1 + i / 100, -.5)
    return Ax(mt=mt, x=x, i=i, g=g, method=method) * np.sqrt(1 + i / 100)


def t_Ax(mt, x, defer=0, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of a deferred whole life insurance, that pays 1 at the end of year of death.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param defer: deferment period
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Deferred Whole Life Insurance (end of year of death)
    """

    return A_x(mt=mt, x=x, x_first=x + 1 + defer, x_last=mt.w + 1, i=i, g=g, method=method)


def t_Ax_(mt, x, defer=0, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of a deferred whole life insurance, that pays 1 at the moment of death.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param defer: deferment period
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Deferred Whole Life Insurance (moment of death)
    """
    return t_Ax(mt=mt, x=x, defer=defer, i=i, g=g, method=method) * np.sqrt(1 + i / 100)


# Term Life Insurance

def nAx(mt, x, n, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of a term life insurance, that pays 1, at the end of the year of death.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: period of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages and periods (udd, cfm, bal)

    :return: Term life insurance (end of year of death)
    """
    if x > mt.w:
        return np.power(1 + i / 100, -1)
    return A_x(mt=mt, x=x, x_first=x + 1, x_last=x + n, i=i, g=g, method=method)


def nAx_(mt, x, n, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of a term life insurance, that pays 1, at the end of the year of death.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: period of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages and periods (udd, cfm, bal)

    :return: Term life insurance (moment of death)
    """
    if x > mt.w:
        return np.power(1 + i / 100, -.5)
    return nAx(mt=mt, x=x, n=n, i=i, g=g, method=method) * np.sqrt(1 + i / 100)


def t_nAx(mt, x, n, defer=0, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of a deferred term life insurance, that pays 1, at the end of the year of death.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: period of the contract
    :param defer: deferment period
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional periods

    :return: Deferred Term Life Insurance (end of year of death).
    """
    return A_x(mt=mt, x=x, x_first=x + 1 + defer, x_last=x + n + defer, i=i, g=g, method=method)


def t_nAx_(mt, x, n, defer=0, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of a deferred term life insurance, that pays 1, at the moment of death.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: period of the contract
    :param defer: deferment period
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional periods

    :return: Deferred Term Life Insurance (moment of death).
    """
    return t_nAx(mt=mt, x=x, n=n, defer=defer, i=i, g=g, method=method) * np.sqrt(1 + i / 100)


# Endowment Insurance

def nAEx(mt, x, n, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of an Endowment life insurance, that pays 1, at the end of year of death
    or 1 if (x) survives to age x+n.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: period of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages and periods (udd, cfm, bal)

    :return: Endowment insurance (end of year of death)
    """
    return nAx(mt=mt, x=x, n=n, i=i, g=g, method=method) + annuities.nEx(mt=mt, x=x, i=i, g=g, n=n, method=method)


def nAEx_(mt, x, n, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of an Endowment life insurance, that pays 1, at the moment of death
    or 1 if (x) survives to age x+n.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: period of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages and periods (udd, cfm, bal)

    :return: Endowment insurance (moment of death)
    """
    return nAx_(mt=mt, x=x, n=n, i=i, g=g, method=method) + annuities.nEx(mt=mt, x=x, i=i, g=g, n=n, method=method)


def t_nAEx(mt, x, n, defer=0, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of a deferred Endowment life insurance that pays 1
    at the end of year of death or 1 if (x) survives to age x+t+n.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: period of the contract
    :param defer: deferment period
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages and periods (udd, cfm, bal)

    :return: Deferred Endowment insurance (end of the year of death)
    """
    return A_x(mt=mt, x=x, x_first=x + 1 + defer, x_last=x + n + defer, i=i, g=g, method=method) + \
        annuities.nEx(mt=mt, x=x, i=i, g=g, n=n + defer, method=method)


def t_nAEx_(mt, x, n, defer=0, i=None, g=.0, method='udd'):
    """
    Returns the Expected Present Value of a deferred Endowment life insurance that pays 1
    at the moment of death or 1 if (x) survives to age x+t+n.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: period of the contract
    :param defer: deferment period
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param method: the method to approximate the fractional ages and periods (udd, cfm, bal)

    :return: Deferred Endowment insurance (moment of death)
    """
    return A_x(mt=mt, x=x, x_first=x + 1 + defer, x_last=x + n + defer, i=i, g=g, method=method) * np.sqrt(
        1 + i / 100) + \
        annuities.nEx(mt=mt, x=x, i=i, g=g, n=n + defer, method=method)


# Life Insurance with linear increment


def IA_x(mt, x, x_first, x_last, i=None, first=1., inc=1., method='udd'):
    """
    Returns the Expected Present Value of a Term Life Insurance , that pays 1+k, at the end of year, if
    death occurs between ages x+k and x+k+1, for k=0, 1, ...
    The capital of the first year equals the rate of the progression.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param x_first: age of first payment
    :param x_last: age of final payment
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param first: amount of the first payment
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)
    :return: Life Insurance with Linear Increment
    """
    if x_first < x: return np.nan
    if x_last < x_first == x: return np.nan
    if x == x_first == x_last: return 0
    if first + (x_last - x_first) * inc < 0: return np.nan
    i = i / 100
    i = float(1 / (1 + i))
    number_of_payments = int((x_last - x_first) + 1)
    if number_of_payments < 1: return 1.  # according to the mortality table x is going to die before x+1
    payments_instants = np.linspace(x_first - x, x_last - x, number_of_payments)
    pppp=[((t - (x_first - x)) * inc + first) for t in payments_instants]


    instalments = [mt.npx(x, n=t - 1, method=method) * mt.nqx(x + t - 1, n=1, method=method) * np.power(i, t) *
                   ((t - (x_first - x)) * inc + first) for t in payments_instants]
    instalments = np.array(instalments)
    return np.sum(instalments)

# Whole LIfe

def IAx(mt, x, i=None, inc=1., method='udd'):
    """
    Returns the Expected Present Value of a Whole Life Insurance , that pays 1+k, at the end of year of death, if
    death occurs between ages x+k and x+k+1, for k=0, 1,...
    The capital of the first year equals the rate of the progression.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Whole Life Insurance with Linear Increment (end of year of death)
    """

    return IA_x(mt=mt, x=x, x_first=x + 1, x_last=mt.w + 1, i=i, first=inc, inc=inc, method=method)


def IAx_(mt, x, i=None, inc=1., method='udd'):
    """
    Returns the Expected Present Value of a Whole Life Insurance , that pays 1+k, at the moment of death, if
    death occurs between ages x+k and x+k+1, for k=0, 1,...
    The capital of the first year equals the rate of the progression.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Whole Life Insurance with Linear Increment (moment of death)
    """

    return IAx(mt=mt, x=x, i=i, inc=inc, method=method) * np.sqrt(1 + i / 100)


def t_IAx(mt, x, defer=0, i=None, inc=1., method='udd'):
    """
    Returns the Expected Present Value of a Whole Life Insurance , that pays 1+k, at the end of year of death, if
    death occurs between ages x+k and x+k+1, for k=0, 1,...
    The capital of the first year equals the rate of the progression.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param defer: deferment period
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages and periods (udd, cfm, bal)

    :return: Deferred Whole Life Insurance with Linear Increment (end of year of death)
    """

    return IA_x(mt=mt, x=x, x_first=x + 1 + defer, x_last=mt.w + 1, i=i, first=inc, inc=inc, method=method)


def t_IAx_(mt, x, defer=0, i=None, inc=1., method='udd'):
    """
    Returns the Expected Present Value of a Whole Life Insurance , that pays 1+k, at the moment of death, if
    death occurs between ages x+k and x+k+1, for k=0, 1,...
    The capital of the first year equals the rate of the progression.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param defer: deferment period
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages and periods (udd, cfm, bal)

    :return: Deferred Whole Life Insurance with Linear Increment (moment of death)
    """

    return t_IAx(mt=mt, x=x, defer=defer, i=i, inc=inc, method=method) * np.sqrt(1 + i / 100)

# Term Life

def nIAx(mt, x, n, i=None, inc=1., method='udd'):
    """
    Returns the Expected Present Value of a Term Life Insurance , that pays 1+k, at the end of year of death, if
    death occurs between ages x+k and x+k+1, for k=0, 1,..., n-1
    The capital of the first year equals the rate of the progression.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Term Life Insurance with Linear Increment (end of year of death)
    """

    return IA_x(mt=mt, x=x, x_first=x + 1, x_last=x + n, i=i, first=inc, inc=inc, method=method)


def nIAx_(mt, x, n, i=None, inc=1., method='udd'):
    """
    Returns the Expected Present Value of a Term Life Insurance , that pays 1+k, at the moment of death, if
    death occurs between ages x+k and x+k+1, for k=0, 1,..., n-1
    The capital of the first year equals the rate of the progression.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Term Life Insurance with Linear Increment (moment of death)
    """

    return nIAx(mt=mt, x=x, n=n, i=i, inc=inc, method=method) * np.sqrt(1 + i / 100)



def t_nIAx(mt, x, n, defer=0, i=None, inc=1., method='udd'):
    """
    Returns the Expected Present Value of a Term Life Insurance , that pays 1+k, at the end of the year of death, if
    death occurs between ages x+k and x+k+1, for k=0, 1,..., n-1
    The capital of the first year equals the rate of the progression.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param defer: deferment periods
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Deferred Term Life Insurance with Linear Increment (end of the year of death)
    """
    return IA_x(mt=mt, x=x, x_first=x + 1 + defer, x_last=x + n + defer, i=i, first=inc, inc=inc, method=method)


def t_nIAx_(mt, x, n, defer=0, i=None, inc=1., method='udd'):
    """
    Returns the Expected Present Value of a Term Life Insurance , that pays 1+k, at the moment of death, if
    death occurs between ages x+k and x+k+1, for k=0, 1,..., n-1
    The capital of the first year equals the rate of the progression.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param defer: deferment periods
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Deferred Term Life Insurance with Linear Increment (moment of death)
    """
    return t_nIAx(mt=mt, x=x, n=n, defer=defer, i=i, inc=inc, method=method) * np.sqrt(1 + i / 100)



# First Capital different from Increase Rate
def t_nIArx(mt, x, n, defer=0, i=None, first_amount=1, inc=1, method='udd'):
    """
    Returns the Expected Present Value term life insurance that pays first_amount + k * increase_amount,
    at the end of the year of death, if death occurs between ages x+defer+k and x+k+defer+1,
    for k=0,..., n-1.
    Allows the computation for decreasing capitals.
    The first capital may differ from the increasing/decreasing amount.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param defer: deferment periods
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param first_amount: insured amount in the first year of the contract
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Term Life Insurance with Linear Increment (emd of the year of death)
    """
    return IA_x(mt, x=x, x_first=x + 1 + defer, x_last=x + defer + n, i=i, first=first_amount, inc=inc, method=method)


def t_nIArx_(mt, x, n, defer=0, i=None, first_amount=1, inc=1, method='udd'):
    """
    Returns the Expected Present Value term life insurance that pays first_amount + k * increase_amount,
    at the moment of the year of death, if death occurs between ages x+defer+k and x+k+defer+1,
    for k=0,..., n-1.
    Allows the computation for decreasing capitals.
    The first capital may differ from the increasing/decreasing amount.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param defer: deferment periods
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param first_amount: insured amount in the first year of the contract
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Term Life Insurance with Linear Increment (moment of death)
    """
    return IA_x(mt, x, x_first=x + 1 + defer, x_last=x + defer + n, i=i, first=first_amount, inc=inc, method=method) * \
        np.sqrt(1 + i / 100)


# Endowment Life Insurance with linear increment

def nIAErx(mt, x, n, i=None, first_amount=1, inc=1, method='udd'):
    """
    Returns the Expected Present Value of an Endowment Life Insurance that pays
    (first_amount + k*increase_amount), at the end of the year of death,
    if death happens between ages x+k and x+k+1, for k=0,..., n-1 and a capital of
    first_amount + (n-1)*increase_amount in case of life at the end of the contract.
    Allows the computation for decreasing capitals.
    The first capital may differ from the increasing/decreasing amount.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param first_amount: insured amount in the first year of the contract
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Endowment Life Insurance with Linear Increment (emd of the year of death)
    """
    return t_nIArx(mt=mt, x=x, n=n, defer=0, i=i, first_amount=first_amount, inc=inc, method=method) + \
        annuities.nEx(mt=mt, x=x, i=i, g=0, n=n, method=method)*(first_amount+inc*(n-1))


def nIAErx_(mt, x, n, i=None, first_amount=1, inc=1, method='udd'):
    """
    Returns the Expected Present Value of an Endowment Life Insurance that pays
    (first_amount + k*increase_amount), at the moment of death,
    if death happens between ages x+k and x+k+1, for k=0,..., n-1 and a capital of
    first_amount + (n-1)*increase_amount in case of life at the end of the contract.
    Allows the computation for decreasing capitals.
    The first capital may differ from the increasing/decreasing amount.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param first_amount: insured amount in the first year of the contract
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Endowment Life Insurance with Linear Increment (moment of death)
    """
    return t_nIArx(mt=mt, x=x, n=n, defer=0, i=i, first_amount=first_amount, inc=inc, method=method)\
        * np.sqrt(1 + i / 100) + \
        annuities.nEx(mt=mt, x=x, i=i, g=0, n=n, method=method)*(first_amount+inc*(n-1))


def t_nIAErx(mt, x, n, defer=0, i=None, first_amount=1, inc=1., method='udd'):
    """
    Returns the Expected Present Value of an Endowment Life Insurance that pays
    (first_amount + k*increase_amount), at the end of the year of death,
    if death happens between ages x+k and x+k+1, for k=0,..., n-1 and a capital of
    first_amount + (n-1)*increase_amount in case of life at the end of the contract.
    Allows the computation for decreasing capitals.
    The first capital may differ from the increasing/decreasing amount.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param defer: deferment periods
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param first_amount: insured amount in the first year of the contract
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Deferred Endowment Life Insurance with Linear Increment (emd of the year of death)
    """
    return IA_x(mt=mt, x=x, x_first=x + 1 + defer, x_last=x + n + defer, i=i, first=first_amount, inc=inc, method=method) + \
        annuities.nEx(mt=mt, x=x, i=i, g=0, n=n + defer, method=method)*(first_amount+(n-1)*inc)


def t_nIAErx_(mt, x, n, defer=0, i=None, first_amount=1, inc=1., method='udd'):
    """
    Returns the Expected Present Value of an Endowment Life Insurance that pays
    (first_amount + k*increase_amount), at the moment of death,
    if death happens between ages x+k and x+k+1, for k=0,..., n-1 and a capital of
    first_amount + (n-1)*increase_amount in case of life at the end of the contract.
    Allows the computation for decreasing capitals.
    The first capital may differ from the increasing/decreasing amount.
    :param mt: table for life x
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param defer: deferment periods
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param first_amount: insured amount in the first year of the contract
    :param inc: linear increment in monetary units, e.g., 1 for one monetary unit
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Endowment Life Insurance with Linear Increment (moment year of death)
    """
    return IA_x(mt=mt, x=x, x_first=x + 1 + defer, x_last=x + n + defer, i=i, first=first_amount, inc=inc, method=method) * \
        np.sqrt(1 + i / 100) + annuities.nEx(mt=mt, x=x, i=i, g=0, n=n + defer, method=method)*(first_amount+(n-1)*inc)




__author__ = "PedroCR"

import numpy as np
import pandas as pd


class MortalityTable:
    """
    Instantiates a life table, where you can pass it in the form of the lx, qx or px. Note that the first value is
    the first age considered in the table.
    The life table will be complete, that is, from age 0 to age w, that is, the last age where lx>0.
    """

    def __init__(self, data_type='q', mt=None, perc=100, last_q=1):
        """
        Initializes the MortalityTable class, so we can construct a mortality table with the usual fields.
        :param data_type: Should be "l" for lx, "p" for px and "q" for qx.
        :param mt: Should be "l" for lx, "p" for px and "q" for qx.
        :param perc: The percentage of qx to use, e.g., you should use 50 for 50%.
        :param last_q: The value for qw.
        """
        if data_type not in ('l', 'q', 'p'):
            return
        if not mt:
            return
        self.__data_type = data_type
        self.__methods = ('udd', 'cfm', 'bal')
        self.__mt = mt
        self.__x0 = np.int32(mt[0])
        self.__last_q = last_q
        self.__w = 0
        self.__lx = []
        self.__px = []
        self.__qx = []
        self.__dx = []
        self.__ex = []
        self.__perc = perc
        self.msn = []

        radical = 100000.
        pperc = perc / 100.
        mt = np.array(mt[1:])
        if data_type == 'l':
            if mt[-1] > 0:
                mt = np.append(mt, 0)
            self.__qx = (mt[:-1] - mt[1:]) / mt[:-1] * pperc
            # self.__qx = np.append(np.zeros(self.x0), self.__qx)
        if data_type == 'q':
            self.__qx = mt * pperc
        if data_type == 'p':
            self.__qx = (1 - mt) * pperc

        if self.__last_q == 1 and self.__qx[-1] < 1 - .1e-10:
            self.__qx = np.append(self.__qx, 1)
        if self.__last_q == 0 and self.__qx[-1] > .1e-10:
            self.__qx = np.append(self.__qx, 0)
        self.__qx = np.append(np.zeros(self.x0), self.__qx)

        self.__px = 1 - self.__qx
        self.__lx = np.array([radical] * (len(self.__qx) + 1))
        self.__dx = np.array([-1] * len(self.__qx))
        self.__ex = np.array([-1] * len(self.__qx))
        for idx_p, p in enumerate(self.__px):
            self.__lx[idx_p + 1] = self.__lx[idx_p] * p
        self.__dx = self.__lx[:-1] * self.__qx
        sum_lx = np.array([sum(self.__lx[l:]) for l in range(len(self.__qx))])
        self.__ex = sum_lx[1:] / self.__lx[:-2]
        self.__ex = np.append(self.__ex, 0) + .5
        self.__w = len(self.__lx) - 2

    def __repr__(self):
        return f"{self.__class__.__name__}{self.data_type, self.mt, self.__perc, self.__last_q}"

    # getters and setters
    @property
    def data_type(self):
        return self.__data_type

    @property
    def w(self):
        return self.__w

    @property
    def mt(self):
        return self.__mt

    @property
    def methods(self):
        return self.__methods

    @property
    def x0(self):
        return self.__x0

    @property
    def lx(self):
        return self.__lx

    @property
    def px(self):
        return self.__px

    @property
    def qx(self):
        return self.__qx

    @property
    def dx(self):
        return self.__dx

    @property
    def ex(self):
        return self.__ex

    @property
    def perc(self):
        return self.__perc

    def df_life_table(self):
        data = {'x': np.arange(self.w + 1), 'lx': self.__lx[:-1], 'dx': self.__dx,
                'qx': self.__qx, 'px': self.__px, 'exo': self.__ex}
        df = pd.DataFrame(data)
        df = df.astype({'x': 'int16'})
        return df

    def lx_udd(self, t):
        if t > self.w + 1:
            return 0.
        if t < 0:
            return np.nan
        int_t = np.int32(t)
        frac_t = t - int_t
        if frac_t == 0:
            return self.__lx[int_t]
        else:
            return self.__lx[int_t] * (1 - frac_t) + self.__lx[int_t + 1] * frac_t

    def lx_cfm(self, t):
        if t > self.w + 1:
            return 0.
        if t < 0:
            return np.nan
        int_t = np.int32(t)
        frac_t = t - int_t
        if frac_t == 0:
            return self.__lx[int_t]
        else:
            return self.__lx[int_t] * np.power(self.__lx[int_t + 1] / self.__lx[int_t], frac_t)

    def lx_bal(self, t):
        if t >= self.w:
            return 0.
        if t < 0:
            return np.nan
        int_t = np.int32(t)
        frac_t = t - int_t
        if frac_t == 0:
            return self.__lx[int_t]
        else:
            inv_lx = 1 / self.__lx[int_t] - frac_t * (1 / self.__lx[int_t] - 1 / self.__lx[int_t + 1])
            return 1 / inv_lx

    def get_lx_method(self, x, method='udd'):
        if method not in self.__methods:
            return np.nan
        if x < 0:
            return np.nan
        if x > self.w + 1:
            return 0
        if method == 'udd':
            return self.lx_udd(x)
        elif method == 'cfm':
            return self.lx_cfm(x)
        elif method == 'bal':
            return self.lx_bal(x)
        else:
            return np.nan

    def get_integral_px_method(self, x, method='udd'):
        """
        This function can be used to approximate the life expectancy between to ages, using the interpolation rules implemented.
        :param x: the value of x, that should be integer
        :param method: the chosen method to approximate px
        :return: the integral of tpx in the interval [0,1] that can be used to approximate life expectancy between ages.
        """
        if method not in self.__methods:
            return np.nan
        if x < 0:
            return np.nan
        if x > self.w + 1:
            return 0
        if int(x) != x:
            return np.nan
        x = int(x)

        if method == 'udd':
            return 1 - .5 * self.qx[x]
        elif method == 'cfm':
            if self.px[x] == 0:
                return .0
            return -self.qx[x] / np.log(self.px[x])
        elif method == 'bal':
            if self.px[x] == 0:
                return .0
            return -self.px[x] / self.qx[x] * np.log(self.px[x])
        else:
            return np.nan

    def nqx(self, x, n=1, method='udd'):
        """
        Obtains the probability that a life x dies before x+t
        :param method: the method used to approximate lx for non-integer x's
        :param x: age at beginning
        :param n: period
        :return: probability of x dying before x+t
        """
        if method not in self.__methods:
            return np.nan
        if x < 0:
            return np.nan
        if n <= 0:
            return .0
        if x + n > self.w + 1:
            return self.__qx[-1]
        l_x = self.get_lx_method(x, method)
        l_x_t = self.get_lx_method(x + n, method)
        self.msn.append(f"{n}_q_{x}=1-({l_x_t} / {l_x})")
        if l_x == 0: return 1.
        return 1 - l_x_t / l_x

    def npx(self, x, n=1, method='udd'):
        """
        Obtains the probability that a life x dies before x+t
        :param method: the method used to approximate lx for non-integer x's
        :param x: age at beginning
        :param n: period
        :return: probability of x dying before x+t
        """
        if method not in self.__methods:
            return np.nan
        if x < 0:
            return np.nan
        if n <= 0:
            return 1.
        if x + n > self.w + 1:
            return self.__px[-1]
        l_x = self.get_lx_method(x, method)
        l_x_t = self.get_lx_method(x + n, method)
        self.msn.append(f"{n}_p_{x}={l_x_t} / {l_x}")
        if l_x == 0: return .0
        return l_x_t / l_x

    def t_nqx(self, x, t=1, n=1, method='udd'):
        """
        Obtains the probability that a life x dies survives to age x+t and dies before x+t+n
        :param method: the method used to approximate lx for non-integer x's
        :param x: age at beginning
        :param t: deferment period
        :param n: period
        :return: probability of x dying after age x+t and before x+t+n
        """
        l_x = self.get_lx_method(x, method)
        l_x_t = self.get_lx_method(x + t, method)
        l_x_t_n = self.get_lx_method(x + t + n, method)
        self.msn.append(f"{t}|{n}_q_{x}={t}_p_{x}  {n}_q_{x + t}={l_x_t} / {l_x} ({l_x_t}-{l_x_t_n}) / {l_x_t}")
        if l_x == 0: return 1.
        return (l_x_t - l_x_t_n) / l_x

    def force_qw_0(self):
        """
        forces the last qx to be equal to zero, to state that there are no more decrements after w
        :return: the qx, px and lx adjusted
        """
        self.__qx[-1] = 0
        self.__px[-1] = 1
        self.__lx[-1] = self.__lx[-2:-1][0]
        self.__dx[-1] = 0

    def exn(self, x, n, method='udd'):
        """
        Computes the approximated life expectancy between two ages.
        :param x: Initial age.
        :param n: Period to be considered.
        :param method: The method used to interpolate.
        :return: The approximated life span between x and x+n, considering the interpolation methods implemented to approximate the integral of the survival function.
        """
        if method not in self.__methods:
            return np.nan
        if x < 0:
            return np.nan
        if n <= 0:
            return 1.

        n_max = n
        if x + n > self.w:
            n_max = self.w - x + 1

        if x == int(x):
            next_age = x
        else:
            next_age = int(x) + 1
        to_complete_age = np.round(next_age - x, 6)
        if to_complete_age >= n_max:
            return n_max * self.npx(x, n_max / 2, method)

        if to_complete_age > 0:
            integral1 = to_complete_age * self.npx(x, to_complete_age / 2, method)
        else:
            integral1 = 0

        complete_years = int(n_max - to_complete_age)
        integrals = [self.npx(x, to_complete_age + t, method) *
                     self.get_integral_px_method(next_age + t, method) for t in range(complete_years)]
        final_period = np.round(n_max - to_complete_age - int(n_max - to_complete_age), 6)

        if final_period > 0:
            integral2 = self.npx(x, int(n_max - to_complete_age), method) * \
                        final_period * self.npx(next_age + int(n - to_complete_age), final_period, method)
        else:
            integral2 = 0

        return integral1 + sum(integrals) + integral2



__author__ = "PedroCR"

import numpy as np
import pandas as pd
from lifeActuary.commutation_table import CommutationFunctions


class CommutationFunctionsFrac(CommutationFunctions):
    """
    Instantiates, for a specific mortality table and interest rate, all the usual commutation functions:
    Dx, Nx, Sx, Cx, Mx, Rx.

    :return: the commutation symbols Dx, Nx, Sx, Cx, Mx, Rx.
    """

    def __init__(self, i=None, g=0, data_type='q', mt=None, perc=100,
                 frac=2, method='udd'):
        CommutationFunctions.__init__(self, i, g, data_type, mt, perc, app_cont=False)
        if method not in self.methods:
            return
        if frac <= 0 or not isinstance(frac, int):
            return

        self.__method = method
        self.__frac = frac

        self.__ages = np.linspace(0, self.w + 1, (self.w + 1) * self.frac + 1)

        radical = 100000.

        self.__lx_frac = np.array([self.npx(x=0, n=x, method=self.__method) for x in self.__ages])
        self.__lx_frac *= radical
        self.__px_frac = np.array([self.npx(x=x, n=1 / self.__frac, method=self.__method) for x in self.__ages])
        self.__qx_frac = 1 - self.__px_frac
        self.__dx_frac = self.__lx_frac[:-1] - self.__lx_frac[1:]
        self.__dx_frac = np.append(self.__dx_frac, 0)

        # Commutations Functions
        self.__Dx_frac = self.__lx_frac[:] * np.power(self.d, self.__ages)
        self.__Nx_frac = np.array([np.sum(self.__Dx_frac[x:]) for x in range(len(self.__ages))])
        self.__Sx_frac = np.array([np.sum(self.__Nx_frac[x:]) for x in range(len(self.__ages))])
        self.__Cx_frac = self.dx_frac * np.power(self.d, self.__ages + 1 / self.__frac)
        self.__Mx_frac = np.array([np.sum(self.__Cx_frac[x:]) for x in range(len(self.__ages))])
        self.__Rx_frac = np.array([np.sum(self.__Mx_frac[x:]) for x in range(len(self.__ages))])

    def __repr__(self):
        return f"{self.__class__.__name__}{self.i, self.g, self.data_type, self.mt, self.perc, self.frac, self.method} "

    def df_commutation_table_frac(self):
        data1 = {'x': self.__ages, 'lx': self.__lx_frac[:], 'dx': self.__dx_frac,
                 'qx': self.__qx_frac, 'px': self.__px_frac}
        data2 = {'Dx': self.__Dx_frac, 'Nx': self.__Nx_frac, 'Sx': self.__Sx_frac, 'Cx': self.__Cx_frac,
                 'Mx': self.__Mx_frac, 'Rx': self.__Rx_frac}
        data = {**data1, **data2}
        df = pd.DataFrame(data)
        return df

    # getters and setters
    @property
    def method(self):
        return self.__method

    @method.setter
    def method(self, m):
        self.__method = m

    @property
    def frac(self):
        return self.__frac

    @frac.setter
    def frac(self, f):
        self.__frac = f

    @property
    def ages(self):
        return self.__ages

    @property
    def lx_frac(self):
        return self.__lx_frac

    @property
    def px_frac(self):
        return self.__px_frac

    @property
    def qx_frac(self):
        return self.__qx_frac

    @property
    def dx_frac(self):
        return self.__dx_frac

    @property
    def Dx_frac(self):
        return self.__Dx_frac

    @property
    def Nx_frac(self):
        return self.__Nx_frac

    @property
    def Sx_frac(self):
        return self.__Sx_frac

    @property
    def Cx_frac(self):
        return self.__Cx_frac

    @property
    def Mx_frac(self):
        return self.__Mx_frac

    @property
    def Rx_frac(self):
        return self.__Rx_frac

    def age_to_index(self, age_int, age_frac):
        """
        Allows us to get the index for a specific age and use all the vectors produced by the class to make computations
        using the age .
        """
        parts = np.round(age_frac * self.frac, 5)
        if int(parts) == parts and age_int == int(age_int):
            return int((age_int + age_frac) * self.__frac)
        return np.nan



__author__ = "PedroCR"

from xml.dom import minidom

http_header = 'https://mort.soa.org/ViewTable.aspx?&TableIdentity='


class SoaTable:
    def __init__(self, table_name):
        '''
        Reads a previously downloaded life table, from Society of Actuaries, in the xml format and prepares
        all the information to create a life (mortality table).
        :param table_name: The SOA table, in xml format, to be read.
        '''
        self.table_name = table_name
        self.xmldoc = minidom.parse(table_name)
        self.table_id = self.xmldoc.getElementsByTagName('TableIdentity')[0].childNodes[0].data
        self.url = http_header + self.table_id
        self.name = self.xmldoc.getElementsByTagName('TableName')[0].childNodes[0].data
        self.contentType = self.xmldoc.getElementsByTagName('ContentType')[0].childNodes[0].data
        self.tableReference = self.xmldoc.getElementsByTagName('TableReference')[0].childNodes[0].data
        self.ages = self.xmldoc.getElementsByTagName('Y')
        self.min_age = int(self.ages[0].attributes['t'].value)
        self.max_age = self.min_age + len(self.ages) - 1
        self.table_qx = [float(age.childNodes[0].data) for age in self.ages]
        self.table_qx.insert(0, self.min_age)



__author__ = "PedroCR"

import numpy as np


# life generic annuity 1 head
def annuity_x(mt, x, x_first, x_last, i=None, g=.0, m=1, method='udd'):
    """
    Computes the present value of an annuity that starts paying 1 at age x, increasing by (1+g/100) and stops
    at age x_w, paying (1+g/100)^{x_w-x}
    :param mt: table for life x
    :param x: age x
    :param x_first: age of first payment
    :param x_last: age of final payment
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate quoted
    :param method: the method to approximate the fractional periods
    :return: the actuarial present value
    """
    if x_first < x: return np.nan
    if x_last < x_first == x: return np.nan
    if int(m) != m: return np.nan
    if x == x_first == x_last: return 1
    i = i / 100
    g = g / 100
    d = float((1 + g) / (1 + i))
    number_of_payments = int(np.round((x_last - x_first) * m + 1, 0))
    payments_instants = np.linspace(x_first - x, x_last - x, number_of_payments)
    # payments_instants = np.arange(x_first - x, x_last - x + 1 / m, 1 / m)
    instalments = [mt.npx(x, n=t, method=method) *
                   np.power(d, t) for t in payments_instants]
    instalments = np.array(instalments) / np.power(1 + g, x_first - x) / m
    return np.sum(instalments)


# Life Annuities - 1 head

# Whole Life Annuities

def ax(mt, x, i=None, g=0, m=1, method='udd'):
    """
    Returns the actuarial present value of an immediate whole life annuity of 1 per time period.
    Payments of 1/m are made m times per year at the end of the periods. If g<>0, payments increase by (1+g/100) each period.
    :param mt: mortality table for life x
    :param x: age x
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate quoted
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """
    if x + 1 / m > mt.w: return 0

    return annuity_x(mt=mt, x=x, x_first=x + 1 / m, x_last=mt.w, i=i, g=g, m=m, method=method)


def aax(mt, x, i=None, g=0, m=1, method='udd'):
    """
    Returns the actuarial present value of a whole life annuity due of 1 per time period. The payments of
    1/m are made m times per year at the beginning of the periods. If g<>0, payments increase by
    (1+g/100) each period
    :param mt: mortality table for life x
    :param x: age x
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate quoted
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """
    if x > mt.w: return 1

    return annuity_x(mt=mt, x=x, x_first=x, x_last=mt.w, i=i, g=g, m=m, method=method)


def t_ax(mt, x, i=None, g=0, m=1, defer=0, method='udd'):
    """
    Returns the actuarial present value of an immediate whole life annuity of 1 per time period,
    deferred t periods. The payments of 1/m are made m times per year at the end of the periods.
    If g<>>0, payments increase by (1+g/100) each period.
    :param mt: mortality table for life x
    :param x: age x
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate quoted
    :param defer: deferment period
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """
    if x + 1 / m + defer > mt.w: return 0

    return annuity_x(mt=mt, x=x, x_first=x + 1 / m + defer, x_last=mt.w, i=i, g=g, m=m, method=method)


def t_aax(mt, x, i=None, g=0, m=1, defer=0, method='udd'):
    """
    Returns the actuarial present value of a due whole life annuity of 1 per time period,
    deferred t periods. The payments of 1/m are made m times per year at the beginning of the periods.
    If g<>>0, payments increase by (1+g/100) each period.
    :param mt: mortality table for life x
    :param x: age x
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate quoted
    :param defer: deferment period
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """
    if x + defer > mt.w: return 0

    return annuity_x(mt=mt, x=x, x_first=x + defer, x_last=mt.w, i=i, g=g, m=m, method=method)

# Temporary Life Annuities

def nax(mt, x, n, i=None, g=0, m=1, method='udd'):
    """
    Returns the actuarial present value of an immediate n term life annuity of 1 per time period.
    The payments of 1/m are made m times per year at the end of the periods.
    If g<>0, payments increase by (1+g/100) each period.
    :param mt: mortality table for life x
    :param x: age x
    :param n: number of years of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """
    if x + 1 / m > mt.w: return 0

    return annuity_x(mt=mt, x=x, x_first=x + 1 / m, x_last=x + n, i=i, g=g, m=m, method=method)


def naax(mt, x, n, i=None, g=0, m=1, method='udd'):
    """
    Returns the actuarial present value of a due n term life annuity of 1 per time period.
    The payments of 1/m are made m times per year at the beginning of the periods.
    If g<>0, payments increase by (1+g/100) each period.
    :param mt: mortality table for life x
    :param x: age x
    :param n: number of years of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """
    if x > mt.w: return 1

    return annuity_x(mt=mt, x=x, x_first=x, x_last=x + n - 1 / m, i=i, g=g, m=m, method=method)


def t_nax(mt, x, n, i=None, g=0, m=1, defer=0, method='udd'):
    """
    Returns the actuarial present value of a immediate $n$ term life annuity of 1 per time period, deferred t
    periods. The payments of 1/m are made m times per year at the end of the periods.
    If g<>0, payments increase by (1+g/100) each period.
    :param mt: table for life x
    :param x: age x
    :param n: number of years of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate quoted
    :param defer: deferment period
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """
    if x + 1 / m + defer > mt.w: return 0

    return annuity_x(mt=mt, x=x, x_first=x + 1 / m + defer, x_last=x + n + defer, i=i, g=g, m=m, method=method)


def t_naax(mt, x, n, i=None, g=0, m=1, defer=0, method='udd'):
    """
    Returns the actuarial present value of a due n term life annuity of 1 per time period, deferred t
    periods. The payments of 1/m are made m times per year at the beginning of the periods.
    If g<>0, payments increase by (1+g/100) each period.
    :param mt: table for life x
    :param x: age x
    :param n: number of years of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param m: frequency of payments per unit of interest rate quoted
    :param defer: deferment period
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """
    if x + defer > mt.w: return 0

    return annuity_x(mt=mt, x=x, x_first=x + defer, x_last=x + n + defer - 1 / m, i=i, g=g, m=m, method=method)


# Annuities with terms evolving arithmetically

def nIax(mt, x, n, i=None, m=1, first_amount=1, increase_amount=1, method='udd'):
    """
    Returns the actuarial present value of an immediate n term life annuity with payments evolving in
    arithmetic progression.
    Payments of 1/m are made m times per year at the end of the periods.
    First amount and Increase amount may be different.
    For decreasing life annuities, the increase_amount should be negative.
    :param mt: mortality table
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param i: interest rate, in percentage (e.g. 2 for 2%)
    :param m: frequency of payments per unit of interest rate
    :param first_amount: amount of the first payment
    :param increase_amount: amount of the increase rate
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """

    ann1 = nax(mt=mt, x=x, n=n, i=i, g=0, m=m, method=method)
    lst_ann = [t_nax(mt=mt, x=x, n=n - t, i=i, g=0, m=m, defer=t, method=method) for t in range(1, n)]

    ann = ann1 * first_amount + np.sum(lst_ann) * increase_amount

    return ann


def nIaax(mt, x, n, i=None, m=1, first_amount=1, increase_amount=1, method='udd'):
    """
    Returns the actuarial present value of a due n term life annuity with payments evolving in
    arithmetic progression.
    Payments of 1/m are made m times per year at the beginning of the periods.
    First amount and Increase amount may be different.
    For decreasing life annuities, the increase_amount should be negative.
    :param mt: mortality table
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param i: interest rate, in percentage (e.g. 2 for 2%)
    :param m: frequency of payments per unit of interest rate
    :param first_amount: amount of the first payment
    :param increase_amount: amount of the increase rate
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """

    ann1 = naax(mt=mt, x=x, n=n, i=i, g=0, m=m, method=method)
    lst_ann = [t_naax(mt=mt, x=x, n=n - t, i=i, g=0, m=m, defer=t, method=method) for t in range(1, n)]

    ann = ann1 * first_amount + np.sum(lst_ann) * increase_amount

    return ann


def t_nIax(mt, x, n, i=None, m=1, defer=0, first_amount=1, increase_amount=1, method='udd'):
    """
    Returns the actuarial present value of a deferred immediate n term life annuity with payments evolving in
    arithmetic progression.
    Payments of 1/m are made m times per year at the end of the periods.
    First amount and Increase amount may be different.
    For decreasing life annuities, the increase_amount should be negative.
    :param mt: mortality table
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param i: interest rate, in percentage (e.g. 2 for 2%)
    :param m: frequency of payments per unit of interest rate
    :param defer: number of deferment years
    :param first_amount: amount of the first payment
    :param increase_amount: amount of the increase rate
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """

    ann1 = t_nax(mt=mt, x=x, n=n, i=i, g=0, m=m, defer=defer, method=method)
    lst_ann = [t_nax(mt=mt, x=x, n=n - t, i=i, g=0, m=m, defer=t, method=method) for t in range(1, n)]

    ann = ann1 * first_amount + np.sum(lst_ann) * increase_amount

    return ann


def t_nIaax(mt, x, n, i, m, defer, first_amount, increase_amount, method):
    """
    Returns the actuarial present value of a deferred due n term life annuity with payments evolving in
    arithmetic progression.
    Payments of 1/m are made m times per year at the beginning of the periods.
    First amount and Increase amount may be different.
    For decreasing life annuities, the increase_amount should be negative.
    :param mt: mortality table
    :param x: age at the beginning of the contract
    :param n: number of years of the contract
    :param i: interest rate, in percentage (e.g. 2 for 2%)
    :param m: frequency of payments per unit of interest rate
    :param defer: number of deferment years
    :param first_amount: amount of the first payment
    :param increase_amount: amount of the increase rate
    :param method: the method to approximate the fractional ages (udd, cfm, bal)

    :return: Expected Present Value (EPV) for payments of 1/m
    """

    ann1 = t_naax(mt=mt, x=x, n=n, i=i, g=0, m=m, defer=defer, method=method)
    lst_ann = [t_naax(mt=mt, x=x, n=n - t, i=i, g=0, m=m, defer=t, method=method) for t in range(1, n)]

    ann = ann1 * first_amount + np.sum(lst_ann) * increase_amount

    return ann


def nEx(mt, x, i=None, g=0, n=0, method='udd'):
    """
    Returns the Actuarial Present Value / Pure Endowment / Deferred Capital
    :param mt: mortality table
    :param x: age at the beginning of the contract
    :param i: technical interest rate (flat rate) in percentage, e.g., 2 for 2%
    :param g: growth rate (flat rate) in percentage, e.g., 2 for 2%
    :param n: number of years of the contract
    :param method: the method to approximate the fractional ages and periods (udd, cfm, bal)

    :return: the present value of a pure endowment of 1 at age x+n
    """

    return t_naax(mt=mt, x=x, n=1, i=i, g=g, m=1, defer=n, method=method)


# Present Value of a series of cash-flows
def present_value(mt, age, spot_rates, capital, probs=None):
    """
    Computes the expected present value of a cash-flows, that can be contingent on some probabilities.
    Payments are considered at the end of the period.

    :param mt: mortality table
    :param age: age at the beginning of the contract
    :param spot_rates: vector of interest rates for the considered time periods
    :param capital: vector of cash-flow amounts
    :param probs: vector of probabilities. For using the instantiated actuarial table, introduce probs=None

    :return: the expected present value of a cash-flow, that can be contingent on some probabilities.
    """
    if len(spot_rates) != len(capital):
        return np.nan
    if probs is None and mt is None:
        return np.nan

    probs_ = None

    if probs is None:
        if age is None:
            return np.nan
        else:
            probs_ = [mt.npx(age, n + 1) for n in range(len(capital))]

    if isinstance(probs, list):
        if len(probs) == len(spot_rates):
            probs_ = probs
        else:
            return np.nan

    if isinstance(probs, (float, int)):
        probs_ = [probs] * len(capital)
    discount = 1 + np.array(spot_rates) / 100.
    discount = np.cumprod(1 / discount)

    return sum([p * capital[idx_p] * discount[idx_p] for idx_p, p in enumerate(probs_)])